"""
Provide the Generation API.
"""
from __future__ import annotations

import asyncio
import json
import logging
import os
import shutil
from asyncio import create_task, Task, as_completed, Semaphore, CancelledError, sleep
from collections.abc import AsyncIterator
from contextlib import suppress
from math import floor
from pathlib import Path
from typing import cast, AsyncContextManager, ParamSpec, Callable, Awaitable, Sequence

import aiofiles
from aiofiles.os import makedirs
from aiofiles.threadpool.text import AsyncTextIOWrapper

from betty.app import App
from betty.json.linked_data import LinkedDataDumpable
from betty.json.schema import Schema
from betty.locale import get_display_name
from betty.model import get_entity_type_name, UserFacingEntity, Entity, GeneratedEntityId
from betty.model.ancestry import is_public
from betty.openapi import Specification
from betty.serde.dump import DictDump, Dump
from betty.string import camel_case_to_kebab_case, camel_case_to_snake_case, upper_camel_case_to_lower_camel_case
from betty.job import Context
from betty.warnings import deprecated


@deprecated('This function is deprecated as of Betty 0.3.2, and will be removed in Betty 0.4.x. Instead, use `logging.logging.getLogger(__name__)`.')
def getLogger() -> logging.Logger:
    """
    Get the site generation logger.
    """
    return logging.getLogger(__name__)


class Generator:
    async def generate(self, job_context: GenerationContext) -> None:
        raise NotImplementedError(repr(self))


class GenerationContext(Context):
    def __init__(self, app: App):
        super().__init__()
        self._app = app

    @property
    def app(self) -> App:
        return self._app


async def generate(app: App) -> None:
    """
    Generate a new site.
    """
    logger = logging.getLogger(__name__)
    job_context = GenerationContext(app)

    with suppress(FileNotFoundError):
        await asyncio.to_thread(shutil.rmtree, app.project.configuration.output_directory_path)
    await makedirs(app.project.configuration.output_directory_path, exist_ok=True)
    logger.info(app.localizer._('Generating your site to {output_directory}.').format(output_directory=app.project.configuration.output_directory_path))

    # The static public assets may be overridden depending on the number of locales rendered, so ensure they are
    # generated before anything else.
    await _generate_static_public(app, job_context)

    jobs = [job async for job in _run_jobs(app, job_context)]
    log_job = create_task(_log_jobs_forever(app, jobs))
    for completed_job in as_completed(jobs):
        await completed_job
    log_job.cancel()
    await _log_jobs(app, jobs)

    os.chmod(app.project.configuration.output_directory_path, 0o755)
    for directory_path_str, subdirectory_names, file_names in os.walk(app.project.configuration.output_directory_path):
        directory_path = Path(directory_path_str)
        for subdirectory_name in subdirectory_names:
            os.chmod(directory_path / subdirectory_name, 0o755)
        for file_name in file_names:
            os.chmod(directory_path / file_name, 0o644)


async def _log_jobs(app: App, jobs: Sequence[Task[None]]) -> None:
    total_job_count = len(jobs)
    completed_job_count = len([job for job in jobs if job.done()])
    logging.getLogger(__name__).info(app.localizer._(
        'Generated {completed_job_count} out of {total_job_count} items ({completed_job_percentage}%).').format(
        completed_job_count=completed_job_count,
        total_job_count=total_job_count,
        completed_job_percentage=floor(completed_job_count / (total_job_count / 100)),
    ))


async def _log_jobs_forever(app: App, jobs: Sequence[Task[None]]) -> None:
    with suppress(CancelledError):
        while True:
            await sleep(5)
            await _log_jobs(app, jobs)


_JobP = ParamSpec('_JobP')


def _run_job(semaphore: Semaphore, f: Callable[_JobP, Awaitable[None]], *args: _JobP.args, **kwargs: _JobP.kwargs) -> Task[None]:
    async def _job():
        async with semaphore:
            await f(*args, **kwargs)
    return create_task(_job())


async def _run_jobs(app: App, job_context: GenerationContext) -> AsyncIterator[Task[None]]:
    semaphore = Semaphore(512)
    yield _run_job(semaphore, _generate_dispatch, job_context)
    yield _run_job(semaphore, _generate_sitemap, job_context)
    yield _run_job(semaphore, _generate_json_schema, job_context)
    yield _run_job(semaphore, _generate_openapi, job_context)

    locales = app.project.configuration.locales

    for locale in locales:
        yield _run_job(semaphore, _generate_public, job_context, locale)

    for entity_type in app.entity_types:
        if not issubclass(entity_type, UserFacingEntity):
            continue
        if app.project.configuration.entity_types[entity_type].generate_html_list:
            for locale in locales:
                yield _run_job(semaphore, _generate_entity_type_list_html, job_context, locale, entity_type)
        yield _run_job(semaphore, _generate_entity_type_list_json, job_context, entity_type)
        for entity in app.project.ancestry[entity_type]:
            if isinstance(entity.id, GeneratedEntityId):
                continue

            yield _run_job(semaphore, _generate_entity_json, job_context, entity_type, entity.id)
            if is_public(entity):
                for locale in locales:
                    yield _run_job(semaphore, _generate_entity_html, job_context, locale, entity_type, entity.id)


async def create_file(path: Path) -> AsyncContextManager[AsyncTextIOWrapper]:
    """
    Create the file for a resource.
    """
    await makedirs(path.parent, exist_ok=True)
    return cast(AsyncContextManager[AsyncTextIOWrapper], aiofiles.open(path, 'w', encoding='utf-8'))


async def create_html_resource(path: Path) -> AsyncContextManager[AsyncTextIOWrapper]:
    """
    Create the file for an HTML resource.
    """
    return await create_file(path / 'index.html')


async def create_json_resource(path: Path) -> AsyncContextManager[AsyncTextIOWrapper]:
    """
    Create the file for a JSON resource.
    """
    return await create_file(path / 'index.json')


async def _generate_dispatch(
    job_context: GenerationContext,
) -> None:
    app = job_context.app
    await app.dispatcher.dispatch(Generator)(job_context),


async def _generate_public(
    job_context: GenerationContext,
    locale: str,
) -> None:
    app = job_context.app
    locale_label = get_display_name(locale, app.localizer.locale)
    logging.getLogger(__name__).debug(app.localizer._('Generating localized public files in {locale}...').format(
        locale=locale_label,
    ))
    async for file_path in app.assets.copytree(Path('public') / 'localized', app.project.configuration.localize_www_directory_path(locale)):
        await app.renderer.render_file(
            file_path,
            job_context=job_context,
            localizer=app.localizers[locale],
        )


async def _generate_static_public(
    app: App,
    job_context: Context,
) -> None:
    logging.getLogger(__name__).info(app.localizer._('Generating static public files...'))
    async for file_path in app.assets.copytree(Path('public') / 'static', app.project.configuration.www_directory_path):
        await app.renderer.render_file(
            file_path,
            job_context=job_context,
        )


async def _generate_entity_type_list_html(
    job_context: GenerationContext,
    locale: str,
    entity_type: type[Entity],
) -> None:
    app = job_context.app
    entity_type_name_fs = camel_case_to_kebab_case(get_entity_type_name(entity_type))
    entity_type_path = app.project.configuration.localize_www_directory_path(locale) / entity_type_name_fs
    template = app.jinja2_environment.select_template([
        f'entity/page-list--{entity_type_name_fs}.html.j2',
        'entity/page-list.html.j2',
    ])
    rendered_html = await template.render_async(
        job_context=job_context,
        localizer=app.localizers[locale],
        page_resource=f'/{entity_type_name_fs}/index.html',
        entity_type=entity_type,
        entities=app.project.ancestry[entity_type],
    )
    async with await create_html_resource(entity_type_path) as f:
        await f.write(rendered_html)


async def _generate_entity_type_list_json(
    job_context: GenerationContext,
    entity_type: type[Entity & LinkedDataDumpable],
) -> None:
    app = job_context.app
    entity_type_name = get_entity_type_name(entity_type)
    entity_type_name_fs = camel_case_to_kebab_case(get_entity_type_name(entity_type))
    entity_type_path = app.project.configuration.www_directory_path / entity_type_name_fs
    data: DictDump[Dump] = {
        '$schema': app.static_url_generator.generate(f'schema.json#/definitions/response/{upper_camel_case_to_lower_camel_case(entity_type_name)}Collection', absolute=True),
        'collection': []
    }
    for entity in app.project.ancestry[entity_type]:
        cast(list[str], data['collection']).append(
            app.url_generator.generate(
                entity,
                'application/json',
                absolute=True,
            ))
    rendered_json = json.dumps(data)
    async with await create_json_resource(entity_type_path) as f:
        await f.write(rendered_json)


async def _generate_entity_html(
    job_context: GenerationContext,
    locale: str,
    entity_type: type[Entity],
    entity_id: str,
) -> None:
    app = job_context.app
    entity = app.project.ancestry[entity_type][entity_id]
    entity_type_name_fs = camel_case_to_kebab_case(get_entity_type_name(entity))
    entity_path = app.project.configuration.localize_www_directory_path(locale) / entity_type_name_fs / entity.id
    rendered_html = await app.jinja2_environment.select_template([
        f'entity/page--{entity_type_name_fs}.html.j2',
        'entity/page.html.j2',
    ]).render_async(
        job_context=job_context,
        localizer=app.localizers[locale],
        page_resource=entity,
        entity_type=entity.type,
        entity=entity,
    )
    async with await create_html_resource(entity_path) as f:
        await f.write(rendered_html)


async def _generate_entity_json(
    job_context: GenerationContext,
    entity_type: type[Entity & LinkedDataDumpable],
    entity_id: str,
) -> None:
    app = job_context.app
    entity_type_name_fs = camel_case_to_kebab_case(get_entity_type_name(entity_type))
    entity_path = app.project.configuration.www_directory_path / entity_type_name_fs / entity_id
    entity = cast('Entity & LinkedDataDumpable', app.project.ancestry[entity_type][entity_id])
    rendered_json = json.dumps(await entity.dump_linked_data(app))
    async with await create_json_resource(entity_path) as f:
        await f.write(rendered_json)


async def _generate_sitemap(
    job_context: GenerationContext,
) -> None:
    app = job_context.app
    sitemap_template = app.jinja2_environment.get_template('sitemap.xml.j2')
    sitemaps = []
    sitemap: list[str] = []
    sitemap_length = 0
    sitemaps.append(sitemap)
    for locale in app.project.configuration.locales:
        for entity in app.project.ancestry:
            if isinstance(entity.id, GeneratedEntityId):
                continue
            if not isinstance(entity, UserFacingEntity):
                continue

            sitemap.append(app.url_generator.generate(
                entity,
                absolute=True,
                locale=locale,
                media_type='text/html',
            ))
            sitemap_length += 1

            if sitemap_length == 50_000:
                sitemap = []
                sitemap_length = 0
                sitemaps.append(sitemap)

    sitemaps_urls = []
    for index, sitemap in enumerate(sitemaps):
        sitemaps_urls.append(app.static_url_generator.generate(
            f'sitemap-{index}.xml',
            absolute=True,
        ))
        rendered_sitemap = await sitemap_template.render_async({
            'urls': sitemap,
        })
        async with aiofiles.open(app.project.configuration.www_directory_path / f'sitemap-{index}.xml', 'w') as f:
            await f.write(rendered_sitemap)

    rendered_sitemap_index = await app.jinja2_environment.get_template('sitemap-index.xml.j2').render_async({
        'sitemaps_urls': sitemaps_urls,
    })
    async with aiofiles.open(app.project.configuration.www_directory_path / 'sitemap.xml', 'w') as f:
        await f.write(rendered_sitemap_index)


async def _generate_json_schema(
    job_context: GenerationContext,
) -> None:
    app = job_context.app
    logging.getLogger(__name__).debug(app.localizer._('Generating JSON Schema...'))
    schema = Schema(app)
    rendered_json = json.dumps(await schema.build())
    async with await create_file(app.project.configuration.www_directory_path / 'schema.json') as f:
        await f.write(rendered_json)


async def _generate_openapi(
    job_context: GenerationContext,
) -> None:
    app = job_context.app
    logging.getLogger(__name__).debug(app.localizer._('Generating OpenAPI specification...'))
    api_directory_path = app.project.configuration.www_directory_path / 'api'
    rendered_json = json.dumps(await Specification(app).build())
    async with await create_json_resource(api_directory_path) as f:
        await f.write(rendered_json)


def _get_entity_type_jinja2_name(entity_type_name: str) -> str:
    return camel_case_to_snake_case(entity_type_name).replace('.', '__')



"""Define Betty's core application functionality."""

from __future__ import annotations

import operator
import weakref
from concurrent.futures import Executor, ProcessPoolExecutor
from contextlib import suppress
from functools import reduce
from graphlib import CycleError, TopologicalSorter
from multiprocessing import get_context
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Mapping, Self, final

import aiohttp

from betty import fs
from betty.app.extension import ListExtensions, Extension, Extensions, build_extension_type_graph, \
    CyclicDependencyError, ExtensionDispatcher, ConfigurableExtension, discover_extension_types
from betty.asyncio import sync, wait
from betty.cache import FileCache
from betty.config import Configurable, FileBasedConfiguration
from betty.dispatch import Dispatcher
from betty.fs import FileSystem, ASSETS_DIRECTORY_PATH
from betty.locale import LocalizerRepository, get_data, DEFAULT_LOCALE, Localizer, Str
from betty.model import Entity, EntityTypeProvider
from betty.model.event_type import EventType, EventTypeProvider, Birth, Baptism, Adoption, Death, Funeral, Cremation, \
    Burial, Will, Engagement, Marriage, MarriageAnnouncement, Divorce, DivorceAnnouncement, Residence, Immigration, \
    Emigration, Occupation, Retirement, Correspondence, Confirmation
from betty.project import Project
from betty.render import Renderer, SequentialRenderer
from betty.serde.dump import minimize, void_none, Dump, VoidableDump, Void
from betty.serde.load import AssertionFailed, Fields, Assertions, OptionalField, Asserter

if TYPE_CHECKING:
    from betty.jinja2 import Environment
    from betty.serve import Server
    from betty.url import StaticUrlGenerator, LocalizedUrlGenerator

CONFIGURATION_DIRECTORY_PATH = fs.HOME_DIRECTORY_PATH / 'configuration'


class _AppExtensions(ListExtensions):
    def __init__(self):
        super().__init__([])

    def _update(self, extensions: list[list[Extension]]) -> None:
        self._extensions = extensions


class AppConfiguration(FileBasedConfiguration):
    def __init__(
        self,
        *,
        locale: str | None = None,
    ):
        super().__init__()
        self._locale: str | None = locale

    @property
    def configuration_file_path(self) -> Path:
        return CONFIGURATION_DIRECTORY_PATH / 'app.json'

    @configuration_file_path.setter
    def configuration_file_path(self, __) -> None:
        pass

    @configuration_file_path.deleter
    def configuration_file_path(self) -> None:
        pass

    @property
    def locale(self) -> str | None:
        return self._locale

    @locale.setter
    def locale(self, locale: str) -> None:
        try:
            get_data(locale)
        except ValueError:
            raise AssertionFailed(Str._(
                '"{locale}" is not a valid IETF BCP 47 language tag.',
                locale=locale,
            ))
        self._locale = locale
        self._dispatch_change()

    def update(self, other: Self) -> None:
        self._locale = other._locale
        self._dispatch_change()

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            OptionalField(
                'locale',
                Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'locale')),
        ),
        )(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return minimize({
            'locale': void_none(self.locale)
        }, True)


@final
class App(Configurable[AppConfiguration]):
    def __init__(
        self,
        configuration: AppConfiguration | None = None,
        project: Project | None = None,
    ):
        super().__init__()
        self._started = False
        self._configuration = configuration or AppConfiguration()
        self._configuration.on_change(self._on_locale_change)
        self._assets: FileSystem | None = None
        self._extensions = _AppExtensions()
        self._extensions_initialized = False
        self._localization_initialized = False
        self._localizer: Localizer | None = None
        self._localizers: LocalizerRepository | None = None
        with suppress(FileNotFoundError):
            wait(self.configuration.read())
        self._project = project or Project()
        self.project.configuration.extensions.on_change(self._update_extensions)

        self._dispatcher: ExtensionDispatcher | None = None
        self._entity_types: set[type[Entity]] | None = None
        self._event_types: set[type[EventType]] | None = None
        self._url_generator: LocalizedUrlGenerator | None = None
        self._static_url_generator: StaticUrlGenerator | None = None
        self._jinja2_environment: Environment | None = None
        self._renderer: Renderer | None = None
        self._http_client: aiohttp.ClientSession | None = None
        self._cache: FileCache | None = None
        self._process_pool: Executor | None = None

    @classmethod
    def _unreduce(cls, dumped_app_configuration: VoidableDump, project: Project) -> Self:
        if dumped_app_configuration is Void:
            app_configuration = None
        else:
            app_configuration = AppConfiguration.load(
                dumped_app_configuration,  # type: ignore[arg-type]
            )
        return App(
            app_configuration,
            project,
        )

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        await self.stop()

    async def start(self) -> None:
        if self._started:
            raise RuntimeError('This app has started already.')
        self._started = True

    async def stop(self) -> None:
        self._started = False
        del self.http_client

    def __del__(self) -> None:
        if self._started:
            raise RuntimeError(f'{self} was started, but never stopped.')

    def _on_locale_change(self) -> None:
        del self.localizer
        del self.localizers

    @property
    def project(self) -> Project:
        return self._project

    def discover_extension_types(self) -> set[type[Extension]]:
        return {*discover_extension_types(), *map(type, self._extensions.flatten())}

    @property
    def extensions(self) -> Extensions:
        if not self._extensions_initialized:
            self._extensions_initialized = True
            self._update_extensions()

        return self._extensions

    def _update_extensions(self) -> None:
        extension_types_enabled_in_configuration = set()
        for app_extension_configuration in self.project.configuration.extensions.values():
            if app_extension_configuration.enabled:
                app_extension_configuration.extension_type.enable_requirement().assert_met()
                extension_types_enabled_in_configuration.add(app_extension_configuration.extension_type)

        extension_types_sorter = TopologicalSorter(
            build_extension_type_graph(extension_types_enabled_in_configuration)
        )
        try:
            extension_types_sorter.prepare()
        except CycleError:
            raise CyclicDependencyError([
                app_extension_configuration.extension_type
                for app_extension_configuration
                in self.project.configuration.extensions.values()
            ])

        extensions = []
        while extension_types_sorter.is_active():
            extension_types_batch = extension_types_sorter.get_ready()
            extensions_batch = []
            for extension_type in extension_types_batch:
                if issubclass(extension_type, ConfigurableExtension) and extension_type in self.project.configuration.extensions:
                    extension: Extension = extension_type(self, configuration=self.project.configuration.extensions[extension_type].extension_configuration)
                else:
                    extension = extension_type(self)
                extensions_batch.append(extension)
                extension_types_sorter.done(extension_type)
            extensions.append(extensions_batch)
        self._extensions._update(extensions)
        del self.assets
        del self.localizers
        del self.localizer
        del self.jinja2_environment
        del self.renderer
        del self.entity_types
        del self.event_types

    @property
    def assets(self) -> FileSystem:
        if self._assets is None:
            assets = FileSystem()
            assets.prepend(ASSETS_DIRECTORY_PATH, 'utf-8')
            for extension in self.extensions.flatten():
                extension_assets_directory_path = extension.assets_directory_path()
                if extension_assets_directory_path is not None:
                    assets.prepend(extension_assets_directory_path, 'utf-8')
            assets.prepend(self.project.configuration.assets_directory_path)
            self._assets = assets
        return self._assets

    @assets.deleter
    def assets(self) -> None:
        self._assets = None

    @property
    def dispatcher(self) -> Dispatcher:
        if self._dispatcher is None:
            self._dispatcher = ExtensionDispatcher(self.extensions)

        return self._dispatcher

    @property
    def url_generator(self) -> LocalizedUrlGenerator:
        from betty.url import AppUrlGenerator

        if self._url_generator is None:
            self._url_generator = AppUrlGenerator(self)
        return self._url_generator

    @property
    def static_url_generator(self) -> StaticUrlGenerator:
        from betty.url import StaticPathUrlGenerator

        if self._static_url_generator is None:
            self._static_url_generator = StaticPathUrlGenerator(self.project.configuration)
        return self._static_url_generator

    @property
    def localizer(self) -> Localizer:
        """
        Get the application's localizer.
        """
        if self._localizer is None:
            self._localizer = wait(self.localizers.get_negotiated(self.configuration.locale or DEFAULT_LOCALE))
        return self._localizer

    @localizer.deleter
    def localizer(self) -> None:
        self._localizer = None

    @property
    def localizers(self) -> LocalizerRepository:
        if self._localizers is None:
            self._localizers = LocalizerRepository(self.assets)
        return self._localizers

    @localizers.deleter
    def localizers(self) -> None:
        self._localizers = None

    @property
    def jinja2_environment(self) -> Environment:
        if not self._jinja2_environment:
            from betty.jinja2 import Environment
            self._jinja2_environment = Environment(self)

        return self._jinja2_environment

    @jinja2_environment.deleter
    def jinja2_environment(self) -> None:
        self._jinja2_environment = None

    @property
    def renderer(self) -> Renderer:
        if not self._renderer:
            from betty.jinja2 import Jinja2Renderer

            self._renderer = SequentialRenderer([
                Jinja2Renderer(self.jinja2_environment, self.project.configuration),
            ])

        return self._renderer

    @renderer.deleter
    def renderer(self) -> None:
        self._renderer = None

    @property
    def http_client(self) -> aiohttp.ClientSession:
        if not self._http_client:
            self._http_client = aiohttp.ClientSession(
                connector=aiohttp.TCPConnector(limit_per_host=5),
                headers={
                    'User-Agent': f'Betty (https://github.com/bartfeenstra/betty) on behalf of {self._project.configuration.base_url}{self._project.configuration.root_path}',
                },
            )
            weakref.finalize(self, sync(self._http_client.close))
        return self._http_client

    @http_client.deleter
    def http_client(self) -> None:
        if self._http_client is not None:
            wait(self._http_client.close())
            self._http_client = None

    @property
    def entity_types(self) -> set[type[Entity]]:
        if self._entity_types is None:
            from betty.model.ancestry import Citation, Enclosure, Event, File, Note, Person, PersonName, Presence, Place, Source

            self._entity_types = reduce(operator.or_, wait(self.dispatcher.dispatch(EntityTypeProvider)()), set()) | {
                Citation,
                Enclosure,
                Event,
                File,
                Note,
                Person,
                PersonName,
                Presence,
                Place,
                Source,
            }
        return self._entity_types

    @entity_types.deleter
    def entity_types(self) -> None:
        self._entity_types = None

    @property
    def event_types(self) -> set[type[EventType]]:
        if self._event_types is None:
            self._event_types = set(wait(self.dispatcher.dispatch(EventTypeProvider)())) | {
                Birth,
                Baptism,
                Adoption,
                Death,
                Funeral,
                Cremation,
                Burial,
                Will,
                Engagement,
                Marriage,
                MarriageAnnouncement,
                Divorce,
                DivorceAnnouncement,
                Residence,
                Immigration,
                Emigration,
                Occupation,
                Retirement,
                Correspondence,
                Confirmation,
            }
        return self._event_types

    @event_types.deleter
    def event_types(self) -> None:
        self._event_types = None

    @property
    def servers(self) -> Mapping[str, Server]:
        from betty import serve
        from betty.extension.demo import DemoServer

        return {
            server.name(): server
            for server
            in [
                *(
                    server
                    for extension in self.extensions.flatten()
                    if isinstance(extension, serve.ServerProvider)
                    for server in extension.servers
                ),
                serve.BuiltinAppServer(self),
                DemoServer(),
            ]
        }

    @property
    def cache(self) -> FileCache:
        if self._cache is None:
            self._cache = FileCache(self.localizer, fs.CACHE_DIRECTORY_PATH)
        return self._cache

    @property
    def process_pool(self) -> Executor:
        if self._process_pool is None:
            # Avoid `fork` so as not to start worker processes with unneeded resources.
            # Settle for `spawn` so all environments use the same start method.
            self._process_pool = ProcessPoolExecutor(mp_context=get_context('spawn'))
        return self._process_pool



"""
Provide locale management for the Graphical User Interface.
"""
from __future__ import annotations

from typing import Any

from PyQt6 import QtGui
from PyQt6.QtWidgets import QComboBox, QLabel, QWidget

from betty.app import App
from betty.asyncio import wait
from betty.gui.text import Caption
from betty.locale import negotiate_locale, get_display_name


class LocalizedObject:
    def __init__(self, app: App, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._app = app

    def showEvent(  # type: ignore[misc]
        self: LocalizedObject & QWidget,
        a0: QtGui.QShowEvent | None,
    ) -> None:
        super().showEvent(a0)  # type: ignore[misc]
        self._set_translatables()
        self._app.configuration.on_change(self._set_translatables)

    def _set_translatables(self) -> None:
        pass


class TranslationsLocaleCollector(LocalizedObject):
    def __init__(self, app: App, allowed_locales: set[str]):
        super().__init__(app)
        self._allowed_locales = allowed_locales

        allowed_locale_names: list[tuple[str, str]] = []
        for allowed_locale in allowed_locales:
            locale_name = get_display_name(allowed_locale)
            if locale_name is not None:
                allowed_locale_names.append((
                    allowed_locale,
                    locale_name,
                ))
        allowed_locale_names = sorted(allowed_locale_names, key=lambda x: x[1])

        def _update_configuration_locale() -> None:
            self._app.configuration.locale = self._configuration_locale.currentData()
        self._configuration_locale = QComboBox()
        for i, (locale, locale_name) in enumerate(allowed_locale_names):
            self._configuration_locale.addItem(locale_name, locale)
            if locale == self._app.configuration.locale:
                self._configuration_locale.setCurrentIndex(i)
        self._configuration_locale.currentIndexChanged.connect(_update_configuration_locale)
        self._configuration_locale_label = QLabel()
        self._configuration_locale_caption = Caption()

        self._set_translatables()
        self._app.configuration.on_change(self._set_translatables)

    @property
    def locale(self) -> QComboBox:
        return self._configuration_locale

    @property
    def rows(self) -> list[list[Any]]:
        return [
            [self._configuration_locale_label, self._configuration_locale],
            [self._configuration_locale_caption],
        ]

    def _set_translatables(self) -> None:
        super()._set_translatables()
        localizer = self._app.localizer
        localizers = self._app.localizers
        self._configuration_locale_label.setText(localizer._('Locale'))
        locale = self.locale.currentData()
        if locale:
            translations_locale = negotiate_locale(
                locale,
                list(localizers.locales),
            )
            if translations_locale is None:
                self._configuration_locale_caption.setText(localizer._('There are no translations for {locale_name}.').format(
                    locale_name=get_display_name(locale, localizer.locale),
                ))
            else:
                negotiated_locale_translations_coverage = wait(localizers.coverage(translations_locale))
                if negotiated_locale_translations_coverage[1] == 0:
                    negotiated_locale_translations_coverage_percentage = 0
                else:
                    negotiated_locale_translations_coverage_percentage = round(100 / (negotiated_locale_translations_coverage[1] / negotiated_locale_translations_coverage[0]))
                self._configuration_locale_caption.setText(localizer._('The translations for {locale_name} are {coverage_percentage}%% complete.').format(
                    locale_name=get_display_name(translations_locale, localizer.locale),
                    coverage_percentage=round(negotiated_locale_translations_coverage_percentage)
                ))



"""
Provide utilities for running jobs concurrently.
"""

from __future__ import annotations

import threading


class Context:
    def __init__(self):
        self._claims_lock = threading.Lock()
        self._claimed_job_ids: set[str] = set()

    def claim(self, job_id: str) -> bool:
        with self._claims_lock:
            if job_id in self._claimed_job_ids:
                return False
            self._claimed_job_ids.add(job_id)
        return True



"""
Provide the Dispatch API.
"""
from typing import Any, Sequence, Callable, Awaitable

TargetedDispatcher = Callable[..., Awaitable[Sequence[Any]]]


class Dispatcher:
    def dispatch(self, target_type: type[Any]) -> TargetedDispatcher:
        raise NotImplementedError(repr(self))



"""
Integrate Betty with Docker.
"""
import asyncio
from pathlib import Path
from types import TracebackType

import docker
from docker.models.containers import Container as DockerContainer


class Container:
    _IMAGE_TAG = 'betty-serve'

    def __init__(self, www_directory_path: Path, docker_directory_path: Path, nginx_configuration_file_path: Path):
        self._docker_directory_path = docker_directory_path
        self._nginx_configuration_file_path = nginx_configuration_file_path
        self._www_directory_path = www_directory_path
        self._client = docker.from_env()
        self.__container: DockerContainer | None = None

    async def __aenter__(self) -> None:
        await self.start()

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        await self.stop()

    async def start(self) -> None:
        await asyncio.to_thread(self._start)

    def _start(self) -> None:
        self._client.images.build(path=str(self._docker_directory_path), tag=self._IMAGE_TAG)
        self._container.start()
        self._container.exec_run(['nginx', '-s', 'reload'])

    async def stop(self) -> None:
        await asyncio.to_thread(self._stop)

    def _stop(self) -> None:
        if self._container is not None:
            self._container.stop()

    @property
    def _container(self) -> DockerContainer:
        if self.__container is None:
            self.__container = self._client.containers.create(
                self._IMAGE_TAG,
                auto_remove=True,
                detach=True,
                volumes={
                    self._nginx_configuration_file_path: {
                        'bind': '/etc/nginx/conf.d/betty.conf',
                        'mode': 'ro',
                    },
                    self._www_directory_path: {
                        'bind': '/var/www/betty',
                        'mode': 'ro',
                    },
                },
            )
        return self.__container

    @property
    def ip(self) -> DockerContainer:
        return self._client.api.inspect_container(self._container.id)['NetworkSettings']['Networks']['bridge']['IPAddress']



"""
Provide the Configuration API.
"""
from __future__ import annotations

import inspect
import weakref
from _weakref import ReferenceType
from collections import OrderedDict
from collections.abc import Callable
from contextlib import suppress, chdir
from pathlib import Path
from reprlib import recursive_repr
from tempfile import TemporaryDirectory
from typing import Generic, Iterable, Iterator, SupportsIndex, Hashable, \
    MutableSequence, MutableMapping, TypeVar, Any, Sequence, overload, cast, Self, TypeAlias

import aiofiles
from aiofiles.os import makedirs

from betty.asyncio import wait, sync
from betty.classtools import repr_instance
from betty.functools import slice_to_range
from betty.locale import Str
from betty.serde.dump import Dumpable, Dump, minimize, VoidableDump, Void
from betty.serde.error import SerdeErrorCollection
from betty.serde.format import FormatRepository
from betty.serde.load import Asserter, Assertion, Assertions

T = TypeVar('T')


_ConfigurationListener: TypeAlias = Callable[[], None]
ConfigurationListener: TypeAlias = 'Configuration | _ConfigurationListener'


class Configuration(Dumpable):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._asserter = Asserter()
        self._on_change_listeners: MutableSequence[ReferenceType[_ConfigurationListener]] = []

    def _dispatch_change(self) -> None:
        for listener_reference in self._on_change_listeners:
            listener = listener_reference()
            if listener is None:
                continue
            listener()

    def _prepare_listener(self, listener: ConfigurationListener) -> ReferenceType[_ConfigurationListener]:
        if isinstance(listener, Configuration):
            listener = listener._dispatch_change
        if inspect.ismethod(listener):  # type: ignore[redundant-expr]
            return weakref.WeakMethod(listener)
        return weakref.ref(listener)

    def on_change(self, listener: ConfigurationListener) -> None:
        self._on_change_listeners.append(self._prepare_listener(listener))

    def remove_on_change(self, listener: ConfigurationListener) -> None:
        self._on_change_listeners.append(self._prepare_listener(listener))

    def update(self, other: Self) -> None:
        raise NotImplementedError(repr(self))

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        """
        Load dumped configuration into a new configuration instance.
        """
        raise NotImplementedError(repr(cls))

    @classmethod
    def assert_load(cls: type[ConfigurationT], configuration: ConfigurationT | None = None) -> Assertion[Dump, ConfigurationT]:
        def _assert_load(dump: Dump) -> ConfigurationT:
            return cls.load(dump, configuration)
        _assert_load.__qualname__ = f'{_assert_load.__qualname__} for {cls.__module__}.{cls.__qualname__}.load'
        return _assert_load


ConfigurationT = TypeVar('ConfigurationT', bound=Configuration)


class FileBasedConfiguration(Configuration):
    def __init__(self):
        super().__init__()
        self._project_directory: TemporaryDirectory | None = None  # type: ignore[type-arg]
        self._configuration_file_path: Path | None = None
        self._autowrite = False

    @property
    def autowrite(self) -> bool:
        return self._autowrite

    @autowrite.setter
    def autowrite(self, autowrite: bool) -> None:
        if autowrite:
            if not self._autowrite:
                self.on_change(self._on_change_write)
        else:
            self.remove_on_change(self._on_change_write)
        self._autowrite = autowrite

    @sync
    async def _on_change_write(self) -> None:
        await self.write()

    async def write(self, configuration_file_path: Path | None = None) -> None:
        if configuration_file_path is not None:
            self.configuration_file_path = configuration_file_path

        await self._write(self.configuration_file_path)

    async def _write(self, configuration_file_path: Path) -> None:
        # Change the working directory to allow absolute paths to be turned relative to the configuration file's directory
        # path.
        formats = FormatRepository()
        dump = formats.format_for(configuration_file_path.suffix[1:]).dump(self.dump())
        try:
            async with aiofiles.open(configuration_file_path, mode='w') as f:
                await f.write(dump)
        except FileNotFoundError:
            await makedirs(configuration_file_path.parent)
            await self.write()
        self._configuration_file_path = configuration_file_path

    async def read(self, configuration_file_path: Path | None = None) -> None:
        if configuration_file_path is not None:
            self.configuration_file_path = configuration_file_path

        formats = FormatRepository()
        with SerdeErrorCollection().assert_valid() as errors:
            # Change the working directory to allow relative paths to be resolved against the configuration file's directory
            # path.
            with chdir(self.configuration_file_path.parent):
                async with aiofiles.open(self.configuration_file_path) as f:
                    read_configuration = await f.read()
                with errors.catch(Str.plain(
                    'in {configuration_file_path}',
                    configuration_file_path=str(self.configuration_file_path.resolve()),
                )):
                    loaded_configuration = self.load(
                        formats.format_for(self.configuration_file_path.suffix[1:]).load(read_configuration)
                    )
        self.update(loaded_configuration)

    def __del__(self) -> None:
        if hasattr(self, '_project_directory') and self._project_directory is not None:
            self._project_directory.cleanup()

    @property
    def configuration_file_path(self) -> Path:
        if self._configuration_file_path is None:
            if self._project_directory is None:
                self._project_directory = TemporaryDirectory()
            wait(self._write(Path(self._project_directory.name) / f'{type(self).__name__}.json'))
        return cast(Path, self._configuration_file_path)

    @configuration_file_path.setter
    def configuration_file_path(self, configuration_file_path: Path) -> None:
        if configuration_file_path == self._configuration_file_path:
            return
        formats = FormatRepository()
        formats.format_for(configuration_file_path.suffix[1:])
        self._configuration_file_path = configuration_file_path

    @configuration_file_path.deleter
    def configuration_file_path(self) -> None:
        if self._autowrite:
            raise RuntimeError('Cannot remove the configuration file path while autowrite is enabled.')
        self._configuration_file_path = None


ConfigurationKey: TypeAlias = SupportsIndex | Hashable | type[Any]
ConfigurationKeyT = TypeVar('ConfigurationKeyT', bound=ConfigurationKey)


class ConfigurationCollection(Configuration, Generic[ConfigurationKeyT, ConfigurationT]):
    _configurations: MutableSequence[ConfigurationT] | MutableMapping[ConfigurationKeyT, ConfigurationT]

    def __init__(
        self,
        configurations: Iterable[ConfigurationT] | None = None,
    ):
        super().__init__()
        if configurations is not None:
            self.append(*configurations)

    def __iter__(self) -> Iterator[ConfigurationKeyT] | Iterator[ConfigurationT]:
        raise NotImplementedError(repr(self))

    def __contains__(self, item: Any) -> bool:
        return item in self._configurations

    def __getitem__(self, configuration_key: ConfigurationKeyT) -> ConfigurationT:
        raise NotImplementedError(repr(self))

    def __delitem__(self, configuration_key: ConfigurationKeyT) -> None:
        self.remove(configuration_key)

    def __len__(self) -> int:
        return len(self._configurations)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        if list(self.keys()) != list(other.keys()):
            return False
        if list(self.values()) != list(other.values()):
            return False
        return True

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, configurations=list(self.values()))

    def _remove_without_dispatch(self, *configuration_keys: ConfigurationKeyT) -> None:
        for configuration_key in configuration_keys:
            with suppress(LookupError):
                self._on_remove(self._configurations[configuration_key])  # type: ignore[call-overload]
            del self._configurations[configuration_key]  # type: ignore[call-overload]

    def remove(self, *configuration_keys: ConfigurationKeyT) -> None:
        self._remove_without_dispatch(*configuration_keys)
        self._dispatch_change()

    def _clear_without_dispatch(self) -> None:
        self._remove_without_dispatch(*self.keys())

    def clear(self) -> None:
        self._clear_without_dispatch()
        self._dispatch_change()

    def _on_add(self, configuration: ConfigurationT) -> None:
        configuration.on_change(self)

    def _on_remove(self, configuration: ConfigurationT) -> None:
        configuration.remove_on_change(self)

    def to_index(self, configuration_key: ConfigurationKeyT) -> int:
        raise NotImplementedError(repr(self))

    def to_indices(self, *configuration_keys: ConfigurationKeyT) -> Iterator[int]:
        for configuration_key in configuration_keys:
            yield self.to_index(configuration_key)

    def to_key(self, index: int) -> ConfigurationKeyT:
        raise NotImplementedError(repr(self))

    def to_keys(self, *indices: int | slice) -> Iterator[ConfigurationKeyT]:
        unique_indices = set()
        for index in indices:
            if isinstance(index, slice):
                for slice_index in slice_to_range(index, self._configurations):
                    unique_indices.add(slice_index)
            else:
                unique_indices.add(index)
        for index in sorted(unique_indices):
            yield self.to_key(index)

    @classmethod
    def _item_type(cls) -> type[ConfigurationT]:
        raise NotImplementedError(repr(cls))

    @classmethod
    def _create_default_item(cls, configuration_key: ConfigurationKeyT) -> ConfigurationT:
        raise NotImplementedError(repr(cls))

    def keys(self) -> Iterator[ConfigurationKeyT]:
        raise NotImplementedError(repr(self))

    def values(self) -> Iterator[ConfigurationT]:
        raise NotImplementedError(repr(self))

    def prepend(self, *configurations: ConfigurationT) -> None:
        raise NotImplementedError(repr(self))

    def append(self, *configurations: ConfigurationT) -> None:
        raise NotImplementedError(repr(self))

    def insert(self, index: int, *configurations: ConfigurationT) -> None:
        raise NotImplementedError(repr(self))

    def move_to_beginning(self, *configuration_keys: ConfigurationKeyT) -> None:
        raise NotImplementedError(repr(self))

    def move_towards_beginning(self, *configuration_keys: ConfigurationKeyT) -> None:
        raise NotImplementedError(repr(self))

    def move_to_end(self, *configuration_keys: ConfigurationKeyT) -> None:
        raise NotImplementedError(repr(self))

    def move_towards_end(self, *configuration_keys: ConfigurationKeyT) -> None:
        raise NotImplementedError(repr(self))


class ConfigurationSequence(ConfigurationCollection[int, ConfigurationT], Generic[ConfigurationT]):
    def __init__(
        self,
        configurations: Iterable[ConfigurationT] | None = None,
    ):
        self._configurations: MutableSequence[ConfigurationT] = []
        super().__init__(configurations)

    def to_index(self, configuration_key: int) -> int:
        return configuration_key

    def to_key(self, index: int) -> int:
        return index

    @overload
    def __getitem__(self, configuration_key: int) -> ConfigurationT:
        pass

    @overload
    def __getitem__(self, configuration_key: slice) -> Sequence[ConfigurationT]:
        pass

    def __getitem__(self, configuration_key: int | slice) -> ConfigurationT | Sequence[ConfigurationT]:
        return self._configurations[configuration_key]

    def __iter__(self) -> Iterator[ConfigurationT]:
        return (configuration for configuration in self._configurations)

    def keys(self) -> Iterator[int]:
        return iter(range(0, len(self._configurations)))

    def values(self) -> Iterator[ConfigurationT]:
        yield from self._configurations

    def update(self, other: Self) -> None:
        raise NotImplementedError(repr(self))

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        else:
            configuration._clear_without_dispatch()
        asserter = Asserter()
        with SerdeErrorCollection().assert_valid():
            configuration.append(*asserter.assert_sequence(Assertions(cls._item_type().assert_load()))(dump))
        return configuration

    def dump(self) -> VoidableDump:
        return minimize([
            configuration.dump()
            for configuration in self._configurations
        ])

    def prepend(self, *configurations: ConfigurationT) -> None:
        for configuration in configurations:
            self._on_add(configuration)
            self._configurations.insert(0, configuration)
        self._dispatch_change()

    def append(self, *configurations: ConfigurationT) -> None:
        for configuration in configurations:
            self._on_add(configuration)
            self._configurations.append(configuration)
        self._dispatch_change()

    def insert(self, index: int, *configurations: ConfigurationT) -> None:
        for configuration in reversed(configurations):
            self._on_add(configuration)
            self._configurations.insert(index, configuration)
        self._dispatch_change()

    def move_to_beginning(self, *configuration_keys: int) -> None:
        self.move_to_end(
            *configuration_keys,
            *[
                index
                for index
                in range(0, len(self._configurations))
                if index not in configuration_keys
            ]
        )

    def move_towards_beginning(self, *configuration_keys: int) -> None:
        for index in configuration_keys:
            self._configurations.insert(index - 1, self._configurations.pop(index))
        self._dispatch_change()

    def move_to_end(self, *configuration_keys: int) -> None:
        for index in configuration_keys:
            self._configurations.append(self._configurations[index])
        for index in reversed(configuration_keys):
            self._configurations.pop(index)
        self._dispatch_change()

    def move_towards_end(self, *configuration_keys: int) -> None:
        for index in reversed(configuration_keys):
            self._configurations.insert(index + 1, self._configurations.pop(index))
        self._dispatch_change()


class ConfigurationMapping(ConfigurationCollection[ConfigurationKeyT, ConfigurationT], Generic[ConfigurationKeyT, ConfigurationT]):
    def __init__(
        self,
        configurations: Iterable[ConfigurationT] | None = None,
    ):
        self._configurations: OrderedDict[ConfigurationKeyT, ConfigurationT] = OrderedDict()
        super().__init__(configurations)

    def _minimize_item_dump(self) -> bool:
        return False

    def to_index(self, configuration_key: ConfigurationKeyT) -> int:
        return list(self._configurations.keys()).index(configuration_key)

    def to_key(self, index: int) -> ConfigurationKeyT:
        return list(self._configurations.keys())[index]

    def __getitem__(self, configuration_key: ConfigurationKeyT) -> ConfigurationT:
        try:
            return self._configurations[configuration_key]
        except KeyError:
            self.append(self._create_default_item(configuration_key))
            return self._configurations[configuration_key]

    def __iter__(self) -> Iterator[ConfigurationKeyT]:
        return (configuration_key for configuration_key in self._configurations)

    def _keys_without_scope(self) -> Iterator[ConfigurationKeyT]:
        return (configuration_key for configuration_key in self._configurations.keys())

    def keys(self) -> Iterator[ConfigurationKeyT]:
        return self._keys_without_scope()

    def values(self) -> Iterator[ConfigurationT]:
        yield from self._configurations.values()

    def update(self, other: Self) -> None:
        self.replace(*other.values())

    def replace(self, *values: ConfigurationT) -> None:
        self_keys = list(self.keys())
        other = {
            self._get_key(value): value
            for value in values
        }
        other_values = list(values)
        other_keys = list(map(self._get_key, other_values))

        # Update items that are kept.
        for key in self_keys:
            if key in other_keys:
                self[key].update(other[key])

        # Add items that are new.
        self._append_without_trigger(*(
            other[key]
            for key
            in other_keys
            if key not in self_keys
        ))

        # Remove items that should no longer be present.
        self._remove_without_dispatch(*(
            key
            for key
            in self_keys
            if key not in other_keys
        ))

        # Ensure everything is in the correct order. This will also trigger reactors.
        self.move_to_beginning(*other_keys)

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        dict_dump = asserter.assert_dict()(dump)
        mapping = asserter.assert_mapping(Assertions(cls._item_type().load))({
            key: cls._load_key(value, key)
            for key, value
            in dict_dump.items()
        })
        configuration.replace(*mapping.values())
        return configuration

    def dump(self) -> VoidableDump:
        dump = {}
        for configuration_item in self._configurations.values():
            item_dump = configuration_item.dump()
            if item_dump is not Void:
                item_dump, configuration_key = self._dump_key(item_dump)
                if self._minimize_item_dump():
                    item_dump = minimize(item_dump)
                dump[configuration_key] = item_dump
        return minimize(dump)

    def prepend(self, *configurations: ConfigurationT) -> None:
        for configuration in configurations:
            configuration_key = self._get_key(configuration)
            self._configurations[configuration_key] = configuration
            configuration.on_change(self)
        self.move_to_beginning(*map(self._get_key, configurations))

    def _append_without_trigger(self, *configurations: ConfigurationT) -> None:
        for configuration in configurations:
            configuration_key = self._get_key(configuration)
            self._configurations[configuration_key] = configuration
            configuration.on_change(self)
        self._move_to_end_without_trigger(*map(self._get_key, configurations))

    def append(self, *configurations: ConfigurationT) -> None:
        self._append_without_trigger(*configurations)
        self._dispatch_change()

    def _insert_without_trigger(self, index: int, *configurations: ConfigurationT) -> None:
        current_configuration_keys = list(self._keys_without_scope())
        self._append_without_trigger(*configurations)
        self._move_to_end_without_trigger(
            *current_configuration_keys[0:index],
            *map(self._get_key, configurations),
            *current_configuration_keys[index:]
        )

    def insert(self, index: int, *configurations: ConfigurationT) -> None:
        self._insert_without_trigger(index, *configurations)
        self._dispatch_change()

    def move_to_beginning(self, *configuration_keys: ConfigurationKeyT) -> None:
        for configuration_key in reversed(configuration_keys):
            self._configurations.move_to_end(configuration_key, False)
        self._dispatch_change()

    def move_towards_beginning(self, *configuration_keys: ConfigurationKeyT) -> None:
        self._move_by_offset(-1, *configuration_keys)

    def _move_to_end_without_trigger(self, *configuration_keys: ConfigurationKeyT) -> None:
        for configuration_key in configuration_keys:
            self._configurations.move_to_end(configuration_key)

    def move_to_end(self, *configuration_keys: ConfigurationKeyT) -> None:
        self._move_to_end_without_trigger(*configuration_keys)
        self._dispatch_change()

    def move_towards_end(self, *configuration_keys: ConfigurationKeyT) -> None:
        self._move_by_offset(1, *configuration_keys)

    def _move_by_offset(self, offset: int, *configuration_keys: ConfigurationKeyT) -> None:
        current_configuration_keys = list(self._keys_without_scope())
        indices = list(self.to_indices(*configuration_keys))
        if offset > 0:
            indices.reverse()
        for index in indices:
            self._insert_without_trigger(index + offset, self._configurations.pop(current_configuration_keys[index]))
        self._dispatch_change()

    def _get_key(self, configuration: ConfigurationT) -> ConfigurationKeyT:
        raise NotImplementedError(repr(self))

    @classmethod
    def _load_key(
        cls,
        item_dump: Dump,
        key_dump: str,
    ) -> Dump:
        raise NotImplementedError(repr(cls))

    def _dump_key(self, item_dump: VoidableDump) -> tuple[VoidableDump, str]:
        raise NotImplementedError(repr(self))


class Configurable(Generic[ConfigurationT]):
    _configuration: ConfigurationT

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @property
    def configuration(self) -> ConfigurationT:
        if not hasattr(self, '_configuration'):
            raise RuntimeError(f'{self} has no configuration. {type(self)}.__init__() must ensure it is set.')
        return self._configuration



"""Integrate Betty with PyInstaller."""
import sys
from glob import glob
from pathlib import Path

from PyInstaller.building.api import PYZ, EXE, COLLECT
from PyInstaller.building.build_main import Analysis
from setuptools import find_packages

from betty._package.pyinstaller.hooks import HOOKS_DIRECTORY_PATH
from betty.app import App
from betty.app.extension import discover_extension_types, Extension
from betty.asyncio import gather
from betty.extension.npm import _Npm, build_assets, _NpmBuilder
from betty.fs import ROOT_DIRECTORY_PATH
from betty.project import ExtensionConfiguration


async def _build_assets() -> None:
    npm_builder_extension_types: list[type[_NpmBuilder & Extension]] = [
        extension_type
        for extension_type
        in discover_extension_types()
        if issubclass(extension_type, _NpmBuilder)
    ]
    async with App() as app:
        app.project.configuration.extensions.append(ExtensionConfiguration(_Npm))
        for extension_type in npm_builder_extension_types:
            app.project.configuration.extensions.append(ExtensionConfiguration(extension_type))
        await gather(*([
            build_assets(app.extensions[extension_type])  # type: ignore[arg-type]
            for extension_type
            in npm_builder_extension_types
        ]))


async def a_pyz_exe_coll() -> tuple[Analysis, PYZ, EXE, COLLECT]:
    """
    Build PyInstaller's spec components.
    """
    if sys.platform == 'linux':
        exe_name = 'betty'
    elif sys.platform == 'darwin':
        exe_name = 'betty.app'
    elif sys.platform == 'win32':
        exe_name = 'betty.exe'
    else:
        raise RuntimeError(f'Unsupported platform {sys.platform}.')

    await _build_assets()
    block_cipher = None
    datas = []
    data_file_path_patterns = [
        'betty/assets/**',
        'betty/extension/*/assets/**',
    ]
    for data_file_path_pattern in data_file_path_patterns:
        for data_file_path_str in glob(data_file_path_pattern, recursive=True, root_dir=ROOT_DIRECTORY_PATH):
            data_file_path = Path(data_file_path_str)
            if data_file_path.is_file():
                datas.append((data_file_path_str, str(data_file_path.parent)))
    hiddenimports = [
        *find_packages(
            '.',
            exclude=[
                'betty.tests',
                'betty.tests.*',
            ],
        ),
        'babel.numbers'
    ]
    a = Analysis(
        ['betty/_package/pyinstaller/main.py'],
        pathex=['./'],
        binaries=[],
        datas=datas,
        hiddenimports=hiddenimports,
        hookspath=[str(HOOKS_DIRECTORY_PATH)],
        runtime_hooks=[],
        excludes=[],
        win_no_prefer_redirects=False,
        win_private_assemblies=False,
        cipher=block_cipher,
        noarchive=False,
    )
    pyz = PYZ(
        a.pure,
        a.zipped_data,
        cipher=block_cipher,
    )
    exe = EXE(
        pyz,
        a.binaries,
        a.scripts,
        [],
        name=exe_name,
        exclude_binaries=True,
        debug=False,
        bootloader_ignore_signals=False,
        strip=False,
        upx=True,
        upx_exclude=[],
        console=False,
        icon=str(ROOT_DIRECTORY_PATH / 'betty' / 'assets' / 'public' / 'static' / 'betty.ico'),
    )
    coll = COLLECT(
        exe,
        a.datas,
        a.zipfiles,
        name='betty',
    )
    return a, pyz, exe, coll



"""
Provide a Sphinx plugin to apply string replacements to source code.
"""
from sphinx.application import Sphinx


def render_replacements(app: Sphinx, docname: str, source: list[str]) -> None:
    """
    Handle Sphinx's source-read event to perform string replacements.
    """
    if app.builder.format != 'html':
        return

    for token_name, value in app.env.config['html_context']['betty_replacements'].items():
        source[0] = source[0].replace('{{{ ' + token_name + ' }}}', value)


def setup(app: Sphinx) -> None:
    """
    Implement Sphinx's extension setup.
    """
    app.connect('source-read', render_replacements)



"""
Provide tools to integrate extensions with `npm <https://www.npmjs.com/>`_.

This extension and module are internal.
"""
from __future__ import annotations

import asyncio
import logging
import os
import shutil
import sys
from asyncio import subprocess as aiosubprocess
from contextlib import suppress
from enum import unique, IntFlag, auto
from pathlib import Path
from subprocess import CalledProcessError
from typing import Sequence

from aiofiles.tempfile import TemporaryDirectory

from betty.app.extension import Extension, discover_extension_types
from betty.app.extension.requirement import Requirement, AnyRequirement, AllRequirements
from betty.asyncio import wait
from betty.fs import iterfiles
from betty.locale import Str, DEFAULT_LOCALIZER
from betty.subprocess import run_process


async def npm(
    arguments: Sequence[str],
    cwd: Path | None = None,
) -> aiosubprocess.Process:
    """
    Run an npm command.
    """
    return await run_process(
        ['npm', *arguments],
        cwd=cwd,
        # Use a shell on Windows so subprocess can find the executables it needs (see
        # https://bugs.python.org/issue17023).
        shell=sys.platform.startswith('win32'),
    )


class _NpmRequirement(Requirement):
    def __init__(self, met: bool):
        super().__init__()
        self._met = met
        self._summary = self._met_summary() if met else self._unmet_summary()
        self._details = Str._('npm (https://www.npmjs.com/) must be available for features that require Node.js packages to be installed. Ensure that the `npm` executable is available in your `PATH`.')

    @classmethod
    def _met_summary(cls) -> Str:
        return Str._('`npm` is available')

    @classmethod
    def _unmet_summary(cls) -> Str:
        return Str._('`npm` is not available')

    @classmethod
    def check(cls) -> _NpmRequirement:
        try:
            wait(npm(['--version']))
            logging.getLogger(__name__).debug(cls._met_summary().localize(DEFAULT_LOCALIZER))
            return cls(True)
        except (CalledProcessError, FileNotFoundError):
            logging.getLogger(__name__).debug(cls._unmet_summary().localize(DEFAULT_LOCALIZER))
            return cls(False)

    def is_met(self) -> bool:
        return self._met

    def summary(self) -> Str:
        return self._summary

    def details(self) -> Str:
        return self._details


def is_assets_build_directory_path(path: Path) -> bool:
    """
    Check if the given path is an assets build directory path.
    """
    return path.is_dir() and len(os.listdir(path)) > 0


class _AssetsRequirement(Requirement):
    def __init__(self, extension_types: set[type[_NpmBuilder & Extension]]):
        super().__init__()
        self._extension_types = extension_types
        self._summary = Str._('Pre-built assets')
        self._details: Str
        if not self.is_met():
            extension_names = sorted(
                extension_type.name()
                for extension_type
                in self._extension_types - self._extension_types_with_built_assets
            )
            self._details = Str._(
                'Pre-built assets are unavailable for {extension_names}.',
                extension_names=', '.join(extension_names,
                                          ))
        else:
            self._details = Str.plain('')

    @property
    def _extension_types_with_built_assets(self) -> set[type[_NpmBuilder & Extension]]:
        return {
            extension_type
            for extension_type
            in self._extension_types
            if is_assets_build_directory_path(_get_assets_build_directory_path(extension_type))
        }

    def is_met(self) -> bool:
        return self._extension_types <= self._extension_types_with_built_assets

    def summary(self) -> Str:
        return self._summary

    def details(self) -> Str:
        return self._details


@unique
class _NpmBuilderCacheScope(IntFlag):
    BETTY = auto()
    PROJECT = auto()


class _NpmBuilder:
    async def npm_build(self, working_directory_path: Path, assets_directory_path: Path) -> None:
        raise NotImplementedError(repr(self))

    @classmethod
    def npm_cache_scope(cls) -> _NpmBuilderCacheScope:
        return _NpmBuilderCacheScope.PROJECT


def discover_npm_builders() -> set[type[_NpmBuilder & Extension]]:
    """
    Gather all extensions that are npm builders.
    """
    return {
        extension_type
        for extension_type
        in discover_extension_types()
        if issubclass(extension_type, _NpmBuilder)
    }


def _get_assets_directory_path(extension_type: type[_NpmBuilder & Extension]) -> Path:
    assert issubclass(extension_type, Extension)
    assert issubclass(extension_type, _NpmBuilder)
    assets_directory_path = extension_type.assets_directory_path()
    if not assets_directory_path:
        raise RuntimeError(f'Extension {extension_type} does not have an assets directory.')
    return assets_directory_path / _Npm.name()


def _get_assets_src_directory_path(extension_type: type[_NpmBuilder & Extension]) -> Path:
    return _get_assets_directory_path(extension_type) / 'src'


def _get_assets_build_directory_path(extension_type: type[_NpmBuilder & Extension]) -> Path:
    return _get_assets_directory_path(extension_type) / 'build'


async def build_assets(extension: _NpmBuilder & Extension) -> Path:
    """
    Build the npm assets for an extension.
    """
    assets_directory_path = _get_assets_build_directory_path(type(extension))
    await _build_assets_to_directory_path(extension, assets_directory_path)
    return assets_directory_path


async def _build_assets_to_directory_path(extension: _NpmBuilder & Extension, assets_directory_path: Path) -> None:
    assert isinstance(extension, Extension)
    assert isinstance(extension, _NpmBuilder)
    with suppress(FileNotFoundError):
        await asyncio.to_thread(shutil.rmtree, assets_directory_path)
    os.makedirs(assets_directory_path)
    async with TemporaryDirectory() as working_directory_path_str:
        working_directory_path = Path(working_directory_path_str)
        await extension.npm_build(Path(working_directory_path), assets_directory_path)


class _Npm(Extension):
    _npm_requirement: _NpmRequirement | None = None
    _assets_requirement: _AssetsRequirement | None = None
    _requirement: Requirement | None = None

    @classmethod
    def _ensure_requirement(cls) -> Requirement:
        if cls._requirement is None:
            cls._npm_requirement = _NpmRequirement.check()
            cls._assets_requirement = _AssetsRequirement(discover_npm_builders())
            assert cls._npm_requirement is not None
            assert cls._assets_requirement is not None
            cls._requirement = AnyRequirement(cls._npm_requirement, cls._assets_requirement)
        return cls._requirement

    @classmethod
    def enable_requirement(cls) -> Requirement:
        return AllRequirements(
            cls._ensure_requirement(),
            super().enable_requirement(),
        )

    async def install(self, extension_type: type[_NpmBuilder & Extension], working_directory_path: Path) -> None:
        self._ensure_requirement()
        if self._npm_requirement:
            self._npm_requirement.assert_met()

        await asyncio.to_thread(
            shutil.copytree,
            _get_assets_src_directory_path(extension_type),
            working_directory_path,
            dirs_exist_ok=True,
        )
        async for file_path in iterfiles(working_directory_path):
            await self._app.renderer.render_file(file_path)
        await npm(['install', '--production'], cwd=working_directory_path)

    def _get_cached_assets_build_directory_path(self, extension_type: type[_NpmBuilder & Extension]) -> Path:
        path = self._app.cache.path / self.name() / extension_type.name()
        if extension_type.npm_cache_scope() == _NpmBuilderCacheScope.PROJECT:
            path /= self.app.project.name
        return path

    async def ensure_assets(self, extension: _NpmBuilder & Extension) -> Path:
        assets_build_directory_paths = [
            _get_assets_build_directory_path(type(extension)),
            self._get_cached_assets_build_directory_path(type(extension)),
        ]
        for assets_build_directory_path in assets_build_directory_paths:
            if is_assets_build_directory_path(assets_build_directory_path):
                return assets_build_directory_path

        if self._npm_requirement:
            self._npm_requirement.assert_met()
        return await self._build_cached_assets(extension)

    async def _build_cached_assets(self, extension: _NpmBuilder & Extension) -> Path:
        assets_directory_path = self._get_cached_assets_build_directory_path(type(extension))
        await _build_assets_to_directory_path(extension, assets_directory_path)
        return assets_directory_path



import sys

from betty.app import App
from betty.asyncio import sync
from betty.gui import BettyApplication
from betty.gui.app import WelcomeWindow


@sync
async def main() -> None:
    """
    Launch Betty for PyInstaller builds.
    """
    async with App() as app:
        qapp = BettyApplication([sys.argv[0]], app=app)
        window = WelcomeWindow(app)
        window.show()
        sys.exit(qapp.exec())

if __name__ == "__main__":
    main()



"""
Provide the Cache API.
"""
import asyncio
import logging
import shutil
from contextlib import suppress
from pathlib import Path

from betty.locale import Localizer


class _Cache:
    async def clear(self) -> None:
        raise NotImplementedError


class FileCache(_Cache):
    def __init__(
        self,
        localizer: Localizer,
        cache_directory_path: Path,
        /,
    ):
        self._localizer = localizer
        self._path = cache_directory_path

    @property
    def path(self) -> Path:
        return self._path

    async def clear(self) -> None:
        with suppress(FileNotFoundError):
            await asyncio.to_thread(shutil.rmtree, self.path)
        logging.getLogger(__name__).info(self._localizer._('All caches cleared.'))



"""
Provide Betty's main data model.
"""
from __future__ import annotations

from collections.abc import MutableSequence
from contextlib import suppress
from enum import Enum
from pathlib import Path
from reprlib import recursive_repr
from typing import Iterable, Any, TYPE_CHECKING
from urllib.parse import quote

from geopy import Point
from typing_extensions import deprecated

from betty.classtools import repr_instance
from betty.json.linked_data import LinkedDataDumpable, dump_context, dump_link, add_json_ld
from betty.json.schema import add_property, ref_json_schema
from betty.locale import Localized, Datey, Str, Localizable, ref_datey
from betty.media_type import MediaType
from betty.model import many_to_many, Entity, one_to_many, many_to_one, many_to_one_to_many, \
    MultipleTypesEntityCollection, EntityCollection, UserFacingEntity, EntityTypeAssociationRegistry, \
    GeneratedEntityId, get_entity_type_name
from betty.model.event_type import EventType, UnknownEventType
from betty.serde.dump import DictDump, Dump, dump_default
from betty.string import camel_case_to_kebab_case

if TYPE_CHECKING:
    from betty.app import App


class Privacy(Enum):
    PUBLIC = 1
    PRIVATE = 2
    UNDETERMINED = 3


class HasPrivacy(LinkedDataDumpable):
    def __init__(
        self,
        *args: Any,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        if [privacy, public, private].count(None) < 2:
            raise ValueError(f'Only one of the `privacy`, `public`, and `private` arguments to {type(self)}.__init__() may be given at a time.')
        if privacy is not None:
            self._privacy = privacy
        elif public is True:
            self._privacy = Privacy.PUBLIC
        elif private is True:
            self._privacy = Privacy.PRIVATE
        else:
            self._privacy = Privacy.UNDETERMINED

    @property
    def own_privacy(self) -> Privacy:
        return self._privacy

    def _get_effective_privacy(self) -> Privacy:
        return self.own_privacy

    @property
    def privacy(self) -> Privacy:
        return self._get_effective_privacy()

    @privacy.setter
    def privacy(self, privacy: Privacy) -> None:
        self._privacy = privacy

    @privacy.deleter
    def privacy(self) -> None:
        self.privacy = Privacy.UNDETERMINED

    @property
    def private(self) -> bool:
        return self.privacy is Privacy.PRIVATE

    @private.setter
    def private(self, private: True) -> None:
        self.privacy = Privacy.PRIVATE

    @property
    def public(self) -> bool:
        # Undetermined privacy defaults to public.
        return self.privacy is not Privacy.PRIVATE

    @public.setter
    def public(self, public: True) -> None:
        self.privacy = Privacy.PUBLIC

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump['private'] = self.private
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'private', {
            '$ref': '#/definitions/privacy',
        })
        definitions = dump_default(schema, 'definitions', dict)
        if 'privacy' not in definitions:
            definitions['privacy'] = {
                'type': 'boolean',
                'description': 'Whether this entity is private (true), or public (false).'
            }
        return schema


def is_private(target: Any) -> bool:
    """
    Check if the given target is private.
    """
    if isinstance(target, HasPrivacy):
        return target.private
    return False


def is_public(target: Any) -> bool:
    """
    Check if the given target is public.
    """
    if isinstance(target, HasPrivacy):
        return target.public
    return True


def resolve_privacy(privacy: Privacy | HasPrivacy | None) -> Privacy:
    """
    Resolve the privacy of a value.
    """
    if privacy is None:
        return Privacy.UNDETERMINED
    if isinstance(privacy, Privacy):
        return privacy
    return privacy.privacy


def merge_privacies(*privacies: Privacy | HasPrivacy | None) -> Privacy:
    """
    Merge multiple privacies into one.
    """
    privacies = {
        resolve_privacy(privacy)
        for privacy
        in privacies
    }
    if Privacy.PRIVATE in privacies:
        return Privacy.PRIVATE
    if Privacy.UNDETERMINED in privacies:
        return Privacy.UNDETERMINED
    return Privacy.PUBLIC


class Dated(LinkedDataDumpable):
    def __init__(
        self,
        *args: Any,
        date: Datey | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.date = date

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        if self.date and is_public(self):
            dump['date'] = await self.date.dump_linked_data(app)
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        schema['type'] = 'object'
        schema['additionalProperties'] = False
        add_property(schema, 'date', await ref_datey(schema, app), False)
        return schema


class Described(LinkedDataDumpable):
    def __init__(
        self,
        *args: Any,
        description: str | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.description = description

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        if self.description is not None:
            dump['description'] = self.description
            dump_context(dump, description='description')
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'description', {
            '$ref': '#/definitions/description',
        }, False)
        definitions = dump_default(schema, 'definitions', dict)
        if 'description' not in definitions:
            definitions['description'] = {
                'type': 'string',
            }
        return schema


class HasMediaType(LinkedDataDumpable):
    def __init__(
        self,
        *args: Any,
        media_type: MediaType | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.media_type = media_type

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        if is_public(self):
            if self.media_type is not None:
                dump['mediaType'] = str(self.media_type)
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'mediaType', ref_media_type(schema), False)
        return schema


def ref_media_type(root_schema: DictDump[Dump]) -> DictDump[Dump]:
    """
    Reference the MediaType schema.
    """
    definitions = dump_default(root_schema, 'definitions', dict)
    if 'mediaType' not in definitions:
        definitions['mediaType'] = {
            'type': 'string',
            'description': 'An IANA media type (https://www.iana.org/assignments/media-types/media-types.xhtml).'
        }
    return {
        '$ref': '#/definitions/mediaType',
    }


class Link(HasMediaType, Localized, Described, LinkedDataDumpable):
    url: str
    relationship: str | None
    label: str | None

    def __init__(
        self,
        url: str,
        *,
        relationship: str | None = None,
        label: str | None = None,
        description: str | None = None,
        media_type: MediaType | None = None,
        locale: str | None = None,
    ):
        super().__init__(
            media_type=media_type,
            description=description,
            locale=locale,
        )
        self.url = url
        self.label = label
        self.relationship = relationship

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump['$schema'] = app.static_url_generator.generate('schema.json#/definitions/link', absolute=True)
        dump['url'] = self.url
        if self.label is not None:
            dump['label'] = self.label
        if self.relationship is not None:
            dump['relationship'] = self.relationship
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        schema['type'] = 'object'
        schema['additionalProperties'] = False
        add_json_ld(schema)
        add_property(schema, '$schema', ref_json_schema(schema))
        add_property(schema, 'label', {
            'type': 'string',
            'description': 'The human-readable label, or link text.'
        }, False)
        add_property(schema, 'url', {
            'type': 'string',
            'format': 'uri',
            'description': 'The full URL to the other resource.'
        })
        add_property(schema, 'relationship', {
            'type': 'string',
            'description': 'The relationship between this resource and the link target (https://en.wikipedia.org/wiki/Link_relation).'
        }, False)
        return schema


async def ref_link(root_schema: DictDump[Dump], app: App) -> DictDump[Dump]:
    """
    Reference the Link schema.
    """
    definitions = dump_default(root_schema, 'definitions', dict)
    if 'link' not in definitions:
        definitions['link'] = await Link.linked_data_schema(app)
    return {
        '$ref': '#/definitions/link',
    }


async def ref_link_collection(root_schema: DictDump[Dump], app: App) -> DictDump[Dump]:
    """
    Reference the schema for a collection of Link instances.
    """
    definitions = dump_default(root_schema, 'definitions', dict)
    if 'linkCollection' not in definitions:
        definitions['linkCollection'] = {
            'type': 'array',
            'items': await ref_link(root_schema, app),
        }
    return {
        '$ref': '#/definitions/linkCollection',
    }


class HasLinks(LinkedDataDumpable):
    def __init__(
        self,
        *args: Any,
        links: MutableSequence[Link] | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._links: MutableSequence[Link] = links if links else []

    @property
    def links(self) -> MutableSequence[Link]:
        return self._links

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        await dump_link(
            dump,
            app,
            *(self.links if is_public(self) else ()),
        )
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'links', await ref_link_collection(schema, app))
        return schema


class HasLinksEntity(HasLinks):
    async def dump_linked_data(  # type: ignore[misc]
        self: HasLinksEntity & Entity,
        app: App,
    ) -> DictDump[Dump]:
        dump: DictDump[Dump] = await super().dump_linked_data(app)  # type: ignore[misc]

        if not isinstance(self.id, GeneratedEntityId):
            await dump_link(dump, app, Link(
                app.static_url_generator.generate(f'/{camel_case_to_kebab_case(get_entity_type_name(self.type))}/{self.id}/index.json'),
                relationship='canonical',
                media_type=MediaType('application/ld+json'),
            ))
            if is_public(self):
                await dump_link(dump, app, *(
                    Link(
                        app.url_generator.generate(self, media_type='text/html', locale=locale),
                        relationship='alternate',
                        media_type=MediaType('text/html'),
                        locale=locale,
                    )
                    for locale
                    in app.project.configuration.locales
                ))

        return dump


@many_to_one('entity', 'betty.model.ancestry.HasNotes', 'notes')
class Note(UserFacingEntity, HasPrivacy, HasLinksEntity, Entity):
    entity: HasNotes

    def __init__(
        self,
        text: str,
        *,
        id: str | None = None,
        entity: HasNotes | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        super().__init__(
            id,
            privacy=privacy,
            public=public,
            private=private,
        )
        self._text = text
        if entity is not None:
            self.entity = entity

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Note')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Notes')

    @property
    def text(self) -> str:
        return self._text

    @property
    def label(self) -> Str:
        return Str.plain(self.text)

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump['@type'] = 'https://schema.org/Thing'
        if self.public:
            dump['text'] = self.text
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'text', {
            'type': 'string'
        }, False)
        return schema


@one_to_many('notes', 'betty.model.ancestry.Note', 'entity')
class HasNotes(LinkedDataDumpable):
    def __init__(  # type: ignore[misc]
        self: HasNotes & Entity,
        *args: Any,
        notes: Iterable[Note] | None = None,
        **kwargs: Any,
    ):
        super().__init__(  # type: ignore[misc]
            *args,
            **kwargs,
        )
        if notes is not None:
            self.notes = notes  # type: ignore[assignment]

    @property
    def notes(self) -> EntityCollection[Note]:  # type: ignore[empty-body]
        pass

    @notes.setter
    def notes(self, notes: Iterable[Note]) -> None:
        pass

    @notes.deleter
    def notes(self) -> None:
        pass

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump['notes'] = [
            app.static_url_generator.generate(f'/note/{quote(note.id)}/index.json')
            for note
            in self.notes
            if not isinstance(note.id, GeneratedEntityId)
        ]
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'notes', {
            '$ref': '#/definitions/entity/noteCollection',
        })
        return schema


@many_to_many('citations', 'betty.model.ancestry.Citation', 'facts')
class HasCitations(LinkedDataDumpable):
    def __init__(  # type: ignore[misc]
        self: HasCitations & Entity,
        *args: Any,
        citations: Iterable[Citation] | None = None,
        **kwargs: Any,
    ):
        super().__init__(  # type: ignore[misc]
            *args,
            **kwargs,
        )
        if citations is not None:
            self.citations = citations  # type: ignore[assignment]

    @property
    def citations(self) -> EntityCollection[Citation]:  # type: ignore[empty-body]
        pass

    @citations.setter
    def citations(self, citations: Iterable[Citation]) -> None:
        pass

    @citations.deleter
    def citations(self) -> None:
        pass

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump['citations'] = [
            app.static_url_generator.generate(f'/citation/{quote(citation.id)}/index.json')
            for citation
            in self.citations
            if not isinstance(citation.id, GeneratedEntityId)
        ]
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'citations', {
            '$ref': '#/definitions/entity/citationCollection',
        })
        return schema


@many_to_many('entities', 'betty.model.ancestry.HasFiles', 'files')
class File(Described, HasPrivacy, HasLinksEntity, HasMediaType, HasNotes, HasCitations, UserFacingEntity, Entity):
    def __init__(
        self,
        path: Path,
        *,
        id: str | None = None,
        media_type: MediaType | None = None,
        description: str | None = None,
        notes: Iterable[Note] | None = None,
        citations: Iterable[Citation] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        links: MutableSequence[Link] | None = None,
    ):
        super().__init__(
            id,
            media_type=media_type,
            description=description,
            notes=notes,
            citations=citations,
            privacy=privacy,
            public=public,
            private=private,
            links=links,
        )
        self._path = path

    @property
    def entities(self) -> EntityCollection[Entity]:  # type: ignore[empty-body]
        pass

    @entities.setter
    def entities(self, entities: Iterable[Entity]) -> None:
        pass

    @entities.deleter
    def entities(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('File')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Files')

    @property
    def path(self) -> Path:
        return self._path

    @property
    def label(self) -> Str:
        return Str.plain(self.description) if self.description else super().label

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump['entities'] = [
            app.static_url_generator.generate(f'/{camel_case_to_kebab_case(get_entity_type_name(entity))}/{quote(entity.id)}/index.json')
            for entity
            in self.entities
            if not isinstance(entity.id, GeneratedEntityId)
        ]
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'entities', {
            'type': 'array',
            'items': {
                'type': 'string',
                'format': 'uri'
            }
        })
        return schema


@many_to_many('files', 'betty.model.ancestry.File', 'entities')
class HasFiles:
    def __init__(  # type: ignore[misc]
        self: HasFiles & Entity,
        *args: Any,
        files: Iterable[File] | None = None,
        **kwargs: Any,
    ):
        super().__init__(  # type: ignore[misc]
            *args,
            **kwargs,
        )
        if files is not None:
            self.files = files  # type: ignore[assignment]

    @property
    def files(self) -> EntityCollection[File]:  # type: ignore[empty-body]
        pass

    @files.setter
    def files(self, files: Iterable[File]) -> None:
        pass

    @files.deleter
    def files(self) -> None:
        pass

    @property
    def associated_files(self) -> Iterable[File]:
        return self.files


@many_to_one('contained_by', 'betty.model.ancestry.Source', 'contains')
@one_to_many('contains', 'betty.model.ancestry.Source', 'contained_by')
@one_to_many('citations', 'betty.model.ancestry.Citation', 'source')
class Source(Dated, HasFiles, HasNotes, HasLinksEntity, HasPrivacy, UserFacingEntity, Entity):
    contained_by: Source | None

    def __init__(
        self,
        name: str | None = None,
        *,
        id: str | None = None,
        author: str | None = None,
        publisher: str | None = None,
        contained_by: Source | None = None,
        contains: Iterable[Source] | None = None,
        notes: Iterable[Note] | None = None,
        date: Datey | None = None,
        files: Iterable[File] | None = None,
        links: MutableSequence[Link] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        super().__init__(
            id,
            notes=notes,
            date=date,
            files=files,
            links=links,
            privacy=privacy,
            public=public,
            private=private,
        )
        self.name = name
        self.author = author
        self.publisher = publisher
        if contained_by is not None:
            self.contained_by = contained_by
        if contains is not None:
            self.contains = contains  # type: ignore[assignment]

    def _get_effective_privacy(self) -> Privacy:
        privacy = super()._get_effective_privacy()
        if self.contained_by:
            return merge_privacies(privacy, self.contained_by.privacy)
        return privacy

    @property
    def contains(self) -> EntityCollection[Source]:  # type: ignore[empty-body]
        pass

    @contains.setter
    def contains(self, contains: Iterable[Source]) -> None:
        pass

    @contains.deleter
    def contains(self) -> None:
        pass

    @property
    def citations(self) -> EntityCollection[Citation]:  # type: ignore[empty-body]
        pass

    @citations.setter
    def citations(self, citations: Iterable[Citation]) -> None:
        pass

    @citations.deleter
    def citations(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Source')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Sources')

    @property
    def label(self) -> Str:
        return Str.plain(self.name) if self.name else super().label

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump['@type'] = 'https://schema.org/Thing'
        dump['contains'] = [
            app.static_url_generator.generate(f'/source/{quote(contained.id)}/index.json')
            for contained
            in self.contains
            if not isinstance(contained.id, GeneratedEntityId)
        ]
        dump['citations'] = [
            app.static_url_generator.generate(f'/citation/{quote(citation.id)}/index.json')
            for citation
            in self.citations
            if not isinstance(citation.id, GeneratedEntityId)
        ]
        if self.contained_by is not None and not isinstance(self.contained_by.id, GeneratedEntityId):
            dump['containedBy'] = app.static_url_generator.generate(f'/source/{quote(self.contained_by.id)}/index.json')
        if self.public:
            if self.name is not None:
                dump_context(dump, name='name')
                dump['name'] = self.name
            if self.author is not None:
                dump['author'] = self.author
            if self.publisher is not None:
                dump['publisher'] = self.publisher
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'name', {
            'type': 'string',
        }, False)
        add_property(schema, 'author', {
            'type': 'string',
        }, False)
        add_property(schema, 'publisher', {
            'type': 'string',
        }, False)
        add_property(schema, 'contains', {
            'type': 'array',
            'items': {
                'type': 'string',
                'format': 'uri',
            },
        })
        add_property(schema, 'citations', {
            '$ref': '#/definitions/entity/citationCollection',
        })
        add_property(schema, 'containedBy', {
            'type': 'string',
            'format': 'uri',
        }, False)
        return schema


@deprecated(f'This class is deprecated as of Betty 0.3.2, and will be removed in Betty 0.4.x. No direct replacement is available. Instead, set the privacy for {Source} entities accordingly.')
class AnonymousSource(Source):
    @property  # type: ignore[override]
    def name(self) -> str:
        return 'private'

    @name.setter
    def name(self, _) -> None:
        # This is a no-op as the name is 'hardcoded'.
        pass

    @name.deleter
    def name(self) -> None:
        # This is a no-op as the name is 'hardcoded'.
        pass


@many_to_many('facts', 'betty.model.ancestry.HasCitations', 'citations')
@many_to_one('source', 'betty.model.ancestry.Source', 'citations')
class Citation(Dated, HasFiles, HasPrivacy, HasLinksEntity, UserFacingEntity, Entity):
    def __init__(
        self,
        *,
        id: str | None = None,
        facts: Iterable[HasCitations] | None = None,
        source: Source | None = None,
        location: Str | None = None,
        date: Datey | None = None,
        files: Iterable[File] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        super().__init__(
            id,
            date=date,
            files=files,
            privacy=privacy,
            public=public,
            private=private,
        )
        if facts is not None:
            self.facts = facts  # type: ignore[assignment]
        self.location = location
        self.source = source

    def _get_effective_privacy(self) -> Privacy:
        privacy = super()._get_effective_privacy()
        if self.source:
            return merge_privacies(privacy, self.source.privacy)
        return privacy

    @property
    def facts(self) -> EntityCollection[HasCitations & Entity]:  # type: ignore[empty-body]
        pass

    @facts.setter
    def facts(self, facts: Iterable[HasCitations & Entity]) -> None:
        pass

    @facts.deleter
    def facts(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Citation')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Citations')

    @property
    def label(self) -> Str:
        return self.location or Str.plain('')

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump['@type'] = 'https://schema.org/Thing'
        dump['facts'] = [
            app.static_url_generator.generate(f'/{camel_case_to_kebab_case(get_entity_type_name(fact))}/{quote(fact.id)}/index.json')
            for fact
            in self.facts
            if not isinstance(fact.id, GeneratedEntityId)
        ]
        if self.source is not None and not isinstance(self.source.id, GeneratedEntityId):
            dump['source'] = app.static_url_generator.generate(f'/source/{quote(self.source.id)}/index.json')
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'source', {
            'type': 'string',
            'format': 'uri'
        }, False)
        add_property(schema, 'facts', {
            'type': 'array',
            'items': {
                'type': 'string',
                'format': 'uri'
            }
        })
        return schema


@deprecated(f'This class is deprecated as of Betty 0.3.2, and will be removed in Betty 0.4.x. No direct replacement is available. Instead, set the privacy for {Citation} entities accordingly.')
class AnonymousCitation(Citation):
    @property  # type: ignore[override]
    def location(self) -> Str:
        return Str._("private (in order to protect people's privacy)")

    @location.setter
    def location(self, _) -> None:
        # This is a no-op as the location is 'hardcoded'.
        pass

    @location.deleter
    def location(self) -> None:
        # This is a no-op as the location is 'hardcoded'.
        pass


class PlaceName(Localized, Dated, LinkedDataDumpable):
    def __init__(
        self,
        name: str,
        *,
        locale: str | None = None,
        date: Datey | None = None,
    ):
        super().__init__(
            date=date,
            locale=locale,
        )
        self._name = name

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented  # pragma: no cover
        return self._name == other._name and self.locale == other.locale

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, name=self.name, locale=self.locale)

    def __str__(self) -> str:
        return self._name

    @property
    def name(self) -> str:
        return self._name

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump['name'] = self.name
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'name', {
            'type': 'string'
        })
        return schema


@many_to_one_to_many(
    'betty.model.ancestry.Place',
    'enclosed_by',
    'encloses',
    'enclosed_by',
    'betty.model.ancestry.Place',
    'encloses',
)
class Enclosure(Dated, HasCitations, Entity):
    encloses: Place | None
    enclosed_by: Place | None

    def __init__(
        self,
        encloses: Place | None = None,
        enclosed_by: Place | None = None,
    ):
        super().__init__()
        self.encloses = encloses
        self.enclosed_by = enclosed_by

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Enclosure')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Enclosures')


@one_to_many('events', 'betty.model.ancestry.Event', 'place')
@one_to_many('enclosed_by', 'betty.model.ancestry.Enclosure', 'encloses')
@one_to_many('encloses', 'betty.model.ancestry.Enclosure', 'enclosed_by')
class Place(HasLinksEntity, HasFiles, HasNotes, HasPrivacy, UserFacingEntity, Entity):
    def __init__(
        self,
        *,
        id: str | None = None,
        names: list[PlaceName] | None = None,
        events: Iterable[Event] | None = None,
        enclosed_by: Iterable[Enclosure] | None = None,
        encloses: Iterable[Enclosure] | None = None,
        notes: Iterable[Note] | None = None,
        coordinates: Point | None = None,
        links: MutableSequence[Link] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
    ):
        super().__init__(
            id,
            notes=notes,
            links=links,
            privacy=privacy,
            public=public,
            private=private,
        )
        self._names = [] if names is None else names
        self._coordinates = coordinates
        if events is not None:
            self.events = events  # type: ignore[assignment]
        if enclosed_by is not None:
            self.enclosed_by = enclosed_by  # type: ignore[assignment]
        if encloses is not None:
            self.encloses = encloses  # type: ignore[assignment]

    @property
    def enclosed_by(self) -> EntityCollection[Enclosure]:  # type: ignore[empty-body]
        pass

    @enclosed_by.setter
    def enclosed_by(self, enclosed_by: Iterable[Enclosure]) -> None:
        pass

    @enclosed_by.deleter
    def enclosed_by(self) -> None:
        pass

    @property
    def encloses(self) -> EntityCollection[Enclosure]:  # type: ignore[empty-body]
        pass

    @encloses.setter
    def encloses(self, encloses: Iterable[Enclosure]) -> None:
        pass

    @encloses.deleter
    def encloses(self) -> None:
        pass

    @property
    def events(self) -> EntityCollection[Event]:  # type: ignore[empty-body]
        pass

    @events.setter
    def events(self, events: Iterable[Event]) -> None:
        pass

    @events.deleter
    def events(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Place')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Places')

    @property
    def names(self) -> list[PlaceName]:
        return self._names

    @property
    def coordinates(self) -> Point | None:
        return self._coordinates

    @coordinates.setter
    def coordinates(self, coordinates: Point):
        self._coordinates = coordinates

    @property
    def label(self) -> Str:
        # @todo Negotiate this by locale and date.
        with suppress(IndexError):
            return Str.plain(self.names[0].name)
        return super().label

    @property
    def associated_files(self) -> Iterable[File]:
        yield from self.files
        for event in self.events:
            yield from event.files

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump_context(
            dump,
            names='name',
            events='event',
            enclosedBy='containedInPlace',
            encloses='containsPlace',
        )
        dump['@type'] = 'https://schema.org/Place'
        dump['names'] = [
            await name.dump_linked_data(app)
            for name
            in self.names
        ]
        dump['events'] = [
            app.static_url_generator.generate(f'/event/{quote(event.id)}/index.json')
            for event
            in self.events
            if not isinstance(event.id, GeneratedEntityId)
        ]
        dump['enclosedBy'] = [
            app.static_url_generator.generate(f'/place/{quote(enclosure.enclosed_by.id)}/index.json')
            for enclosure
            in self.enclosed_by
            if enclosure.enclosed_by is not None and not isinstance(enclosure.enclosed_by.id, GeneratedEntityId)
        ]
        dump['encloses'] = [
            app.static_url_generator.generate(f'/place/{quote(enclosure.encloses.id)}/index.json')
            for enclosure
            in self.encloses
            if enclosure.encloses is not None and not isinstance(enclosure.encloses.id, GeneratedEntityId)
        ]
        if self.coordinates is not None:
            dump['coordinates'] = {
                '@type': 'https://schema.org/GeoCoordinates',
                'latitude': self.coordinates.latitude,
                'longitude': self.coordinates.longitude,
            }
            dump_context(dump, coordinates='geo')
            dump_context(
                dump['coordinates'],  # type: ignore[arg-type]
                latitude='latitude',
            )
            dump_context(
                dump['coordinates'],  # type: ignore[arg-type]
                longitude='longitude',
            )
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'names', {
            'type': 'array',
            'items': await PlaceName.linked_data_schema(app),
        })
        add_property(schema, 'enclosedBy', {
            '$ref': '#/definitions/entity/placeCollection'
        }, False)
        add_property(schema, 'encloses', {
            '$ref': '#/definitions/entity/placeCollection'
        })
        coordinate_schema: DictDump[Dump] = {
            'type': 'number',
        }
        coordinates_schema: DictDump[Dump] = {
            'type': 'object',
            'additionalProperties': False,
        }
        add_property(coordinates_schema, 'latitude', coordinate_schema, False)
        add_property(coordinates_schema, 'longitude', coordinate_schema, False)
        add_json_ld(coordinates_schema, schema)
        add_property(schema, 'coordinates', coordinates_schema, False)
        add_property(schema, 'events', {
            '$ref': '#/definitions/entity/eventCollection'
        })
        return schema


class PresenceRole:
    @classmethod
    def name(cls) -> str:
        raise NotImplementedError(repr(cls))

    @property
    def label(self) -> Str:
        raise NotImplementedError(repr(self))


def ref_role(root_schema: DictDump[Dump]) -> DictDump[Dump]:
    """
    Reference the PresenceRole schema.
    """
    definitions = dump_default(root_schema, 'definitions', dict)
    if 'role' not in definitions:
        definitions['role'] = {
            'type': 'string',
            'description': "A person's role in an event.",
        }
    return {
        '$ref': '#/definitions/role',
    }


class Subject(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'subject'

    @property
    def label(self) -> Str:
        return Str._('Subject')


class Witness(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'witness'

    @property
    def label(self) -> Str:
        return Str._('Witness')


class Beneficiary(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'beneficiary'

    @property
    def label(self) -> Str:
        return Str._('Beneficiary')


class Attendee(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'attendee'

    @property
    def label(self) -> Str:
        return Str._('Attendee')


class Speaker(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'speaker'

    @property
    def label(self) -> Str:
        return Str._('Speaker')


class Celebrant(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'celebrant'

    @property
    def label(self) -> Str:
        return Str._('Celebrant')


class Organizer(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'organizer'

    @property
    def label(self) -> Str:
        return Str._('Organizer')


@many_to_one_to_many(
    'betty.model.ancestry.Person',
    'presences',
    'person',
    'event',
    'betty.model.ancestry.Event',
    'presences',
)
class Presence(HasPrivacy, Entity):
    person: Person | None
    event: Event | None
    role: PresenceRole

    def __init__(
        self,
        person: Person | None,
        role: PresenceRole,
        event: Event | None,
    ):
        super().__init__(None)
        self.person = person
        self.role = role
        self.event = event

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Presence')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Presences')

    @property
    def label(self) -> Str:
        return Str._(
            'Presence of {person} at {event}',
            person=self.person.label if self.person else Str._('Unknown'),
            event=self.event.label if self.event else Str._('Unknown'),
        )

    def _get_effective_privacy(self) -> Privacy:
        return merge_privacies(
            super()._get_effective_privacy(),
            self.person,
            self.event,
        )


@many_to_one('place', 'betty.model.ancestry.Place', 'events')
@one_to_many('presences', 'betty.model.ancestry.Presence', 'event')
class Event(Dated, HasFiles, HasCitations, HasNotes, Described, HasPrivacy, HasLinksEntity, UserFacingEntity, Entity):
    place: Place | None

    def __init__(
        self,
        *,
        id: str | None = None,
        event_type: type[EventType] = UnknownEventType,
        date: Datey | None = None,
        files: Iterable[File] | None = None,
        citations: Iterable[Citation] | None = None,
        notes: Iterable[Note] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        place: Place | None = None,
        description: str | None = None,
    ):
        super().__init__(
            id,
            date=date,
            files=files,
            citations=citations,
            notes=notes,
            privacy=privacy,
            public=public,
            private=private,
            description=description,
        )
        self._event_type = event_type
        if place is not None:
            self.place = place

    @property
    def label(self) -> Str:
        format_kwargs: dict[str, str | Localizable] = {
            'event_type': self._event_type.label(),
        }
        subjects = [
            presence.person
            for presence
            in self.presences
            if presence.public and isinstance(presence.role, Subject) and presence.person is not None and presence.person.public
        ]
        if subjects:
            format_kwargs['subjects'] = Str.call(lambda localizer: ', '.join(person.label.localize(localizer) for person in subjects))
        if self.description is not None:
            format_kwargs['event_description'] = self.description

        if subjects:
            if self.description is None:
                return Str._('{event_type} of {subjects}', **format_kwargs)
            else:
                return Str._('{event_type} ({event_description}) of {subjects}', **format_kwargs)
        if self.description is None:
            return Str._('{event_type}', **format_kwargs)
        else:
            return Str._('{event_type} ({event_description})', **format_kwargs)

    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, id=self._id, type=self._event_type)

    @property
    def presences(self) -> EntityCollection[Presence]:  # type: ignore[empty-body]
        pass

    @presences.setter
    def presences(self, presences: Iterable[Presence]) -> None:
        pass

    @presences.deleter
    def presences(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Event')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Events')

    @property
    def event_type(self) -> type[EventType]:
        return self._event_type

    @property
    def associated_files(self) -> Iterable[File]:
        files = [
            *self.files,
            *[file for citation in self.citations for file in citation.associated_files],
        ]
        # Preserve the original order.
        seen = set()
        for file in files:
            if file in seen:
                continue
            seen.add(file)
            yield file

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump_context(dump, presences='performer')
        dump['@type'] = 'https://schema.org/Event'
        dump['type'] = self.event_type.name()
        dump['eventAttendanceMode'] = 'https://schema.org/OfflineEventAttendanceMode'
        dump['eventStatus'] = 'https://schema.org/EventScheduled'
        dump['presences'] = presences = []
        if self.date is not None and self.public:
            await self.date.datey_dump_linked_data(
                dump['date'],  # type: ignore[arg-type]
                'startDate',
                'endDate',
            )
        for presence in self.presences:
            if presence.person and not isinstance(presence.person.id, GeneratedEntityId):
                presences.append(self._dump_event_presence(presence, app))
        if self.place is not None and not isinstance(self.place.id, GeneratedEntityId):
            dump['place'] = app.static_url_generator.generate(f'/place/{quote(self.place.id)}/index.json')
            dump_context(dump, place='location')
        return dump

    def _dump_event_presence(self, presence: Presence, app: App) -> DictDump[Dump]:
        assert presence.person
        dump: DictDump[Dump] = {
            '@type': 'https://schema.org/Person',
            'person': app.static_url_generator.generate(f'/person/{quote(presence.person.id)}/index.json'),
        }
        if presence.public:
            dump['role'] = presence.role.name()
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'type', {
            'type': 'string',
        })
        add_property(schema, 'place', {
            'type': 'string',
            'format': 'uri',
        }, False)
        presence_schema: DictDump[Dump] = {
            'type': 'object',
            'additionalProperties': False,
        }
        add_property(presence_schema, 'role', ref_role(schema), False)
        add_property(presence_schema, 'person', {
            'type': 'string',
            'format': 'uri',
        })
        add_json_ld(presence_schema, schema)
        add_property(schema, 'presences', {
            'type': 'array',
            'items': presence_schema,
        })
        add_property(schema, 'eventStatus', {
            'type': 'string',
        })
        add_property(schema, 'eventAttendanceMode', {
            'type': 'string',
        })
        return schema


@many_to_one('person', 'betty.model.ancestry.Person', 'names')
class PersonName(Localized, HasCitations, HasPrivacy, Entity):
    person: Person | None

    def __init__(
        self,
        *,
        id: str | None = None,
        person: Person | None = None,
        individual: str | None = None,
        affiliation: str | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        locale: str | None = None,
    ):
        if not individual and not affiliation:
            raise ValueError('The individual and affiliation names must not both be empty.')
        super().__init__(
            id,
            privacy=privacy,
            public=public,
            private=private,
            locale=locale,
        )
        self._individual = individual
        self._affiliation = affiliation
        # Set the person association last, because the association requires comparisons, and self.__eq__() uses the
        # individual and affiliation names.
        self.person = person

    def _get_effective_privacy(self) -> Privacy:
        privacy = super()._get_effective_privacy()
        if self.person:
            return merge_privacies(privacy, self.person.privacy)
        return privacy

    def __repr__(self) -> str:
        return repr_instance(self, id=self.id, individual=self.individual, affiliation=self.affiliation)

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Person name')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('Person names')

    @property
    def individual(self) -> str | None:
        return self._individual

    @property
    def affiliation(self) -> str | None:
        return self._affiliation

    @property
    def label(self) -> Str:
        return Str._(
            '{individual_name} {affiliation_name}',
            individual_name='…' if not self.individual else self.individual,
            affiliation_name='…' if not self.affiliation else self.affiliation,
        )

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        if self.public:
            if self.individual is not None:
                dump_context(dump, individual='givenName')
                dump['individual'] = self.individual
            if self.affiliation is not None:
                dump_context(dump, affiliation='familyName')
                dump['affiliation'] = self.affiliation
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'individual', {
            'type': 'string',
        }, False)
        add_property(schema, 'affiliation', {
            'type': 'string',
        }, False)
        return schema


@many_to_many('parents', 'betty.model.ancestry.Person', 'children')
@many_to_many('children', 'betty.model.ancestry.Person', 'parents')
@one_to_many('presences', 'betty.model.ancestry.Presence', 'person')
@one_to_many('names', 'betty.model.ancestry.PersonName', 'person')
class Person(HasFiles, HasCitations, HasNotes, HasLinksEntity, HasPrivacy, UserFacingEntity, Entity):
    def __init__(
        self,
        *,
        id: str | None = None,
        files: Iterable[File] | None = None,
        citations: Iterable[Citation] | None = None,
        links: MutableSequence[Link] | None = None,
        notes: Iterable[Note] | None = None,
        privacy: Privacy | None = None,
        public: bool | None = None,
        private: bool | None = None,
        parents: Iterable[Person] | None = None,
        children: Iterable[Person] | None = None,
        presences: Iterable[Presence] | None = None,
        names: Iterable[PersonName] | None = None,
    ):
        super().__init__(
            id,
            files=files,
            citations=citations,
            links=links,
            notes=notes,
            privacy=privacy,
            public=public,
            private=private,
        )
        if children is not None:
            self.children = children  # type: ignore[assignment]
        if parents is not None:
            self.parents = parents  # type: ignore[assignment]
        if presences is not None:
            self.presences = presences  # type: ignore[assignment]
        if names is not None:
            self.names = names  # type: ignore[assignment]

    @property
    def parents(self) -> EntityCollection[Person]:  # type: ignore[empty-body]
        pass

    @parents.setter
    def parents(self, parents: Iterable[Person]) -> None:
        pass

    @parents.deleter
    def parents(self) -> None:
        pass

    @property
    def children(self) -> EntityCollection[Person]:  # type: ignore[empty-body]
        pass

    @children.setter
    def children(self, children: Iterable[Person]) -> None:
        pass

    @children.deleter
    def children(self) -> None:
        pass

    @property
    def presences(self) -> EntityCollection[Presence]:  # type: ignore[empty-body]
        pass

    @presences.setter
    def presences(self, presences: Iterable[Presence]) -> None:
        pass

    @presences.deleter
    def presences(self) -> None:
        pass

    @property
    def names(self) -> EntityCollection[PersonName]:  # type: ignore[empty-body]
        pass

    @names.setter
    def names(self, names: Iterable[PersonName]) -> None:
        pass

    @names.deleter
    def names(self) -> None:
        pass

    @classmethod
    def entity_type_label(cls) -> Str:
        return Str._('Person')

    @classmethod
    def entity_type_label_plural(cls) -> Str:
        return Str._('People')

    @property
    def siblings(self) -> list[Person]:
        siblings = []
        for parent in self.parents:
            for sibling in parent.children:
                if sibling != self and sibling not in siblings:
                    siblings.append(sibling)
        return siblings

    @property
    def associated_files(self) -> Iterable[File]:
        files = [
            *self.files,
            *[
                file
                for name
                in self.names
                for citation
                in name.citations
                for file
                in citation.associated_files
            ],
            *[
                file
                for presence
                in self.presences
                if presence.event is not None
                for file
                in presence.event.associated_files
            ]
        ]
        # Preserve the original order.
        seen = set()
        for file in files:
            if file in seen:
                continue
            seen.add(file)
            yield file

    @property
    def label(self) -> Str:
        for name in self.names:
            if name.public:
                return name.label
        return super().label

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        dump_context(
            dump,
            names='name',
            parents='parent',
            children='child',
            siblings='sibling',
        )
        dump['@type'] = 'https://schema.org/Person'
        dump['parents'] = [
            app.static_url_generator.generate(f'/person/{quote(parent.id)}/index.json')
            for parent
            in self.parents
            if not isinstance(parent.id, GeneratedEntityId)
        ]
        dump['children'] = [
            app.static_url_generator.generate(f'/person/{quote(child.id)}/index.json')
            for child
            in self.children
            if not isinstance(child.id, GeneratedEntityId)
        ]
        dump['siblings'] = [
            app.static_url_generator.generate(f'/person/{quote(sibling.id)}/index.json')
            for sibling
            in self.siblings
            if not isinstance(sibling.id, GeneratedEntityId)
        ]
        dump['presences'] = [
            self._dump_person_presence(presence, app)
            for presence
            in self.presences
            if presence.event is not None and not isinstance(presence.event.id, GeneratedEntityId)
        ]
        if self.public:
            dump['names'] = [
                await name.dump_linked_data(app)
                for name
                in self.names
                if name.public
            ]
        else:
            dump['names'] = []
        return dump

    def _dump_person_presence(self, presence: Presence, app: App) -> DictDump[Dump]:
        assert presence.event
        dump: DictDump[Dump] = {
            'event': app.static_url_generator.generate(f'/event/{quote(presence.event.id)}/index.json'),
        }
        dump_context(dump, event='performerIn')
        if presence.public:
            dump['role'] = presence.role.name()
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        add_property(schema, 'names', {
            'type': 'array',
            'items': await PersonName.linked_data_schema(app),
        })
        add_property(schema, 'parents', {
            '$ref': '#/definitions/entity/personCollection',
        })
        add_property(schema, 'children', {
            '$ref': '#/definitions/entity/personCollection',
        })
        add_property(schema, 'siblings', {
            '$ref': '#/definitions/entity/personCollection',
        })
        presence_schema: DictDump[Dump] = {
            'type': 'object',
            'additionalProperties': False,
        }
        add_property(presence_schema, 'role', ref_role(schema), False)
        add_property(presence_schema, 'event', {
            'type': 'string',
            'format': 'uri',
        })
        add_json_ld(presence_schema, schema)
        add_property(schema, 'presences', {
            'type': 'array',
            'items': presence_schema,
        })
        return schema


class Ancestry(MultipleTypesEntityCollection[Entity]):
    def __init__(self):
        super().__init__()
        self._check_graph = True

    def add_unchecked_graph(self, *entities: Entity) -> None:
        self._check_graph = False
        try:
            self.add(*entities)
        finally:
            self._check_graph = True

    def _on_add(self, *entities: Entity) -> None:
        super()._on_add(*entities)
        if self._check_graph:
            self.add(*self._get_associates(*entities))

    def _get_associates(self, *entities: Entity) -> Iterable[Entity]:
        for entity in entities:
            for association in EntityTypeAssociationRegistry.get_all_associations(entity):
                for associate in EntityTypeAssociationRegistry.get_associates(entity, association):
                    yield associate



"""
Provide rendering utilities using `Jinja2 <https://jinja.palletsprojects.com>`_.
"""
from __future__ import annotations

import datetime
from collections import defaultdict
from pathlib import Path
from threading import Lock
from typing import Callable, Any, cast, \
    Mapping, TypeVar

import aiofiles
from aiofiles import os as aiofiles_os
from jinja2 import Environment as Jinja2Environment, select_autoescape, FileSystemLoader, pass_context, \
    Template as Jinja2Template
from jinja2.runtime import StrictUndefined, Context, DebugUndefined, new_context

from betty.app import App
from betty.html import CssProvider, JsProvider
from betty.jinja2.filter import FILTERS
from betty.jinja2.test import TESTS
from betty.job import Context as JobContext
from betty.locale import Date, Localizer, \
    DEFAULT_LOCALIZER
from betty.model import Entity, get_entity_type
from betty.model.ancestry import Citation
from betty.project import ProjectConfiguration
from betty.render import Renderer
from betty.serde.dump import Dumpable, DictDump, VoidableDump, Void, Dump

T = TypeVar('T')


def context_app(context: Context) -> App:
    """
    Get the current app from the Jinja2 context.
    """
    return cast(Environment, context.environment).app


def context_job_context(context: Context) -> JobContext | None:
    """
    Get the current job context from the Jinja2 context.
    """
    job_context = context.resolve_or_missing('job_context')
    return job_context if isinstance(job_context, JobContext) else None


def context_localizer(context: Context) -> Localizer:
    """
    Get the current localizer from the Jinja2 context.
    """
    localizer = context.resolve_or_missing('localizer')
    if isinstance(localizer, Localizer):
        return localizer
    raise RuntimeError('No `localizer` context variable exists in this Jinja2 template.')


class _Citer:
    __slots__ = '_lock', '_cited'

    def __init__(self):
        self._lock = Lock()
        self._cited: list[Citation] = []

    def __iter__(self) -> enumerate[Citation]:
        return enumerate(self._cited, 1)

    def __len__(self) -> int:
        return len(self._cited)

    def cite(self, citation: Citation) -> int:
        with self._lock:
            if citation not in self._cited:
                self._cited.append(citation)
            return self._cited.index(citation) + 1


class _Breadcrumb(Dumpable):
    def __init__(self, label: str, url: str):
        self._label = label
        self._url = url

    def dump(self) -> DictDump[Dump]:
        return {
            '@type': 'ListItem',
            'name': self._label,
            'item': self._url,
        }


class _Breadcrumbs(Dumpable):
    def __init__(self):
        self._breadcrumbs: list[_Breadcrumb] = []

    def append(self, label: str, url: str) -> None:
        self._breadcrumbs.append(_Breadcrumb(label, url))

    def dump(self) -> VoidableDump:
        if not self._breadcrumbs:
            return Void
        return {
            '@context': 'https://schema.org',
            '@type': 'BreadcrumbList',
            'itemListElement': [
                {
                    'position': position,
                    **breadcrumb.dump(),
                }
                for position, breadcrumb
                in enumerate(self._breadcrumbs, 1)
            ],
        }


class EntityContexts:
    def __init__(self, *entities: Entity) -> None:
        self._contexts: dict[type[Entity], Entity | None] = defaultdict(lambda: None)
        for entity in entities:
            self._contexts[entity.type] = entity

    def __getitem__(self, entity_type_or_type_name: type[Entity] | str) -> Entity | None:
        if isinstance(entity_type_or_type_name, str):
            entity_type = get_entity_type(entity_type_or_type_name)
        else:
            entity_type = entity_type_or_type_name
        return self._contexts[entity_type]

    def __call__(self, *entities: Entity) -> EntityContexts:
        updated_contexts = EntityContexts()
        for entity in entities:
            updated_contexts._contexts[entity.type] = entity
        return updated_contexts


class Jinja2Provider:
    @property
    def globals(self) -> dict[str, Any]:
        return {}

    @property
    def filters(self) -> dict[str, Callable[..., Any]]:
        return {}

    @property
    def tests(self) -> dict[str, Callable[..., bool]]:
        return {}


class Template(Jinja2Template):
    environment: Environment

    def new_context(
        self,
        vars: dict[str, Any] | None = None,
        shared: bool = False,
        locals: Mapping[str, Any] | None = None,
    ) -> Context:
        return new_context(
            self.environment,
            self.name,
            self.blocks,
            vars,
            shared,
            {
                'citer': _Citer(),
                'breadcrumbs': _Breadcrumbs(),
                **self.globals,
            },
            locals,
        )


class Environment(Jinja2Environment):
    template_class = Template
    globals: dict[str, Any]
    filters: dict[str, Callable[..., Any]]
    tests: dict[str, Callable[..., bool]]

    def __init__(self, app: App):
        template_directory_paths = [str(path / 'templates') for path, _ in app.assets.paths]
        super().__init__(
            loader=FileSystemLoader(template_directory_paths),
            auto_reload=app.project.configuration.debug,
            enable_async=True,
            undefined=DebugUndefined if app.project.configuration.debug else StrictUndefined,
            autoescape=select_autoescape(['html.j2']),
            trim_blocks=True,
            lstrip_blocks=True,
            extensions=[
                'jinja2.ext.do',
                'jinja2.ext.i18n',
            ],
        )

        self.app = app

        if app.project.configuration.debug:
            self.add_extension('jinja2.ext.debug')

        self._init_i18n()
        self._init_globals()
        self.filters.update(FILTERS)
        self.tests.update(TESTS)
        self._init_extensions()

    def _init_i18n(self) -> None:
        self.install_gettext_callables(  # type: ignore[attr-defined]
            gettext=self._gettext,
            ngettext=self._ngettext,
            pgettext=self._pgettext,
            npgettext=self._npgettext,
        )
        self.policies['ext.i18n.trimmed'] = True

    @pass_context
    def _gettext(self, context: Context, message: str) -> str:
        return context_localizer(context).gettext(message)

    @pass_context
    def _ngettext(self, context: Context, message_singular: str, message_plural: str, n: int) -> str:
        return context_localizer(context).ngettext(message_singular, message_plural, n)

    @pass_context
    def _pgettext(self, context: Context, gettext_context: str, message: str) -> str:
        return context_localizer(context).pgettext(gettext_context, message)

    @pass_context
    def _npgettext(self, context: Context, gettext_context: str, message_singular: str, message_plural: str, n: int) -> str:
        return context_localizer(context).npgettext(gettext_context, message_singular, message_plural, n)

    def _init_globals(self) -> None:
        self.globals['app'] = self.app
        today = datetime.date.today()
        self.globals['today'] = Date(today.year, today.month, today.day)
        # Ideally we would use the Dispatcher for this. However, it is asynchronous only.
        self.globals['public_css_paths'] = [
            path
            for extension in self.app.extensions.flatten()
            if isinstance(extension, CssProvider)
            for path in extension.public_css_paths
        ]
        self.globals['public_js_paths'] = [
            path
            for extension in self.app.extensions.flatten()
            if isinstance(extension, JsProvider)
            for path in extension.public_js_paths
        ]
        self.globals['entity_contexts'] = EntityContexts()
        self.globals['localizer'] = DEFAULT_LOCALIZER

    def _init_extensions(self) -> None:
        for extension in self.app.extensions.flatten():
            if isinstance(extension, Jinja2Provider):
                self.globals.update(extension.globals)
                self.filters.update(extension.filters)
                self.tests.update(extension.tests)


Template.environment_class = Environment


class Jinja2Renderer(Renderer):
    def __init__(self, environment: Environment, configuration: ProjectConfiguration):
        self._environment = environment
        self._configuration = configuration

    @property
    def file_extensions(self) -> set[str]:
        return {'.j2'}

    async def render_file(
        self,
        file_path: Path,
        *,
        job_context: JobContext | None = None,
        localizer: Localizer | None = None,
    ) -> Path:
        destination_file_path = file_path.parent / file_path.stem
        data: dict[str, Any] = {}
        if job_context is not None:
            data['job_context'] = job_context
        if localizer is not None:
            data['localizer'] = localizer
        try:
            relative_file_destination_path = destination_file_path.relative_to(self._configuration.www_directory_path)
        except ValueError:
            pass
        else:
            resource = '/'.join(relative_file_destination_path.parts)
            if self._configuration.locales.multilingual:
                resource_parts = resource.lstrip('/').split('/')
                if resource_parts[0] in map(lambda x: x.alias, self._configuration.locales.values()):
                    resource = '/'.join(resource_parts[1:])
            data['page_resource'] = resource
        async with aiofiles.open(file_path) as f:
            template_source = await f.read()
        rendered = await self._environment.from_string(template_source, self._environment.globals).render_async(data)
        async with aiofiles.open(destination_file_path, 'w', encoding='utf-8') as f:
            await f.write(rendered)
        await aiofiles_os.remove(file_path)
        return destination_file_path



"""
Provide the desktop application/Graphical User Interface.
"""
import webbrowser
from datetime import datetime
from os import path
from pathlib import Path
from urllib.parse import urlencode

from PyQt6.QtCore import Qt, QCoreApplication, QObject
from PyQt6.QtGui import QIcon, QAction
from PyQt6.QtWidgets import QFormLayout, QWidget, QVBoxLayout, QHBoxLayout, QFileDialog, QPushButton

from betty import about
from betty.about import report
from betty.app import App
from betty.asyncio import sync, wait
from betty.gui import get_configuration_file_filter
from betty.gui.error import ExceptionCatcher
from betty.gui.locale import TranslationsLocaleCollector
from betty.gui.serve import ServeDemoWindow, ServeDocsWindow
from betty.gui.text import Text
from betty.gui.window import BettyMainWindow
from betty.locale import Str, Localizable
from betty.project import ProjectConfiguration


class BettyPrimaryWindow(BettyMainWindow):
    def __init__(
        self,
        app: App,
        /,
    ):
        super().__init__(app)
        self.setWindowIcon(QIcon(path.join(path.dirname(__file__), 'assets', 'public', 'static', 'betty-512x512.png')))

        menu_bar = self.menuBar()
        assert menu_bar is not None

        betty_menu = menu_bar.addMenu('&Betty')
        assert betty_menu is not None
        self.betty_menu = betty_menu

        self.new_project_action = QAction(self)
        self.new_project_action.setShortcut('Ctrl+N')
        self.new_project_action.triggered.connect(
            lambda _: self.new_project(),
        )
        betty_menu.addAction(self.new_project_action)

        self.open_project_action = QAction(self)
        self.open_project_action.setShortcut('Ctrl+O')
        self.open_project_action.triggered.connect(
            lambda _: self.open_project(),
        )
        betty_menu.addAction(self.open_project_action)

        self._demo_action = QAction(self)
        self._demo_action.triggered.connect(
            lambda _: self._demo(),
        )
        betty_menu.addAction(self._demo_action)

        self.open_application_configuration_action = QAction(self)
        self.open_application_configuration_action.triggered.connect(
            lambda _: self.open_application_configuration(),
        )
        betty_menu.addAction(self.open_application_configuration_action)

        self.clear_caches_action = QAction(self)
        self.clear_caches_action.triggered.connect(
            lambda _: self.clear_caches(),
        )
        betty_menu.addAction(self.clear_caches_action)

        self.exit_action = QAction(self)
        self.exit_action.setShortcut('Ctrl+Q')
        self.exit_action.triggered.connect(QCoreApplication.quit)
        betty_menu.addAction(self.exit_action)

        help_menu = menu_bar.addMenu('')
        assert help_menu is not None
        self.help_menu = help_menu

        self.report_bug_action = QAction(self)
        self.report_bug_action.triggered.connect(
            lambda _: self.report_bug(),
        )
        help_menu.addAction(self.report_bug_action)

        self.request_feature_action = QAction(self)
        self.request_feature_action.triggered.connect(
            lambda _: self.request_feature(),
        )
        help_menu.addAction(self.request_feature_action)

        self._docs_action = QAction(self)
        self._docs_action.triggered.connect(
            lambda _: self._docs(),
        )
        help_menu.addAction(self._docs_action)

        self.about_action = QAction(self)
        self.about_action.triggered.connect(
            lambda _: self._about_betty(),
        )
        help_menu.addAction(self.about_action)

    @property
    def window_title(self) -> Localizable:
        return Str.plain('Betty')

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self.new_project_action.setText(self._app.localizer._('New project...'))
        self.open_project_action.setText(self._app.localizer._('Open project...'))
        self._demo_action.setText(self._app.localizer._('View demo site...'))
        self.open_application_configuration_action.setText(self._app.localizer._('Settings...'))
        self.clear_caches_action.setText(self._app.localizer._('Clear all caches'))
        self.exit_action.setText(self._app.localizer._('Exit'))
        self.help_menu.setTitle('&' + self._app.localizer._('Help'))
        self.report_bug_action.setText(self._app.localizer._('Report a bug'))
        self.request_feature_action.setText(self._app.localizer._('Request a new feature'))
        self._docs_action.setText(self._app.localizer._('View documentation'))
        self.about_action.setText(self._app.localizer._('About Betty'))

    def report_bug(self) -> None:
        with ExceptionCatcher(self):
            body = f'''
## Summary

## Steps to reproduce

## Expected behavior

## System report
```
{report()}
```
'''.strip()
            webbrowser.open_new_tab('https://github.com/bartfeenstra/betty/issues/new?' + urlencode({
                'body': body,
                'labels': 'bug',
            }))

    def request_feature(self) -> None:
        with ExceptionCatcher(self):
            body = '''
## Summary

## Expected behavior

'''.strip()
            webbrowser.open_new_tab('https://github.com/bartfeenstra/betty/issues/new?' + urlencode({
                'body': body,
                'labels': 'enhancement',
            }))

    def _docs(self) -> None:
        with ExceptionCatcher(self):
            serve_window = ServeDocsWindow(self._app, parent=self)
            serve_window.show()

    def _about_betty(self) -> None:
        with ExceptionCatcher(self):
            about_window = _AboutBettyWindow(self._app, parent=self)
            about_window.show()

    def open_project(self) -> None:
        with ExceptionCatcher(self):
            from betty.gui.project import ProjectWindow

            configuration_file_path_str, __ = QFileDialog.getOpenFileName(
                self,
                self._app.localizer._('Open your project from...'),
                '',
                get_configuration_file_filter().localize(self._app.localizer),
            )
            if not configuration_file_path_str:
                return
            wait(self._app.project.configuration.read(Path(configuration_file_path_str)))
            project_window = ProjectWindow(self._app)
            project_window.show()
            self.close()

    def new_project(self) -> None:
        with ExceptionCatcher(self):
            from betty.gui.project import ProjectWindow

            configuration_file_path_str, __ = QFileDialog.getSaveFileName(
                self,
                self._app.localizer._('Save your new project to...'),
                '',
                get_configuration_file_filter().localize(self._app.localizer),
            )
            if not configuration_file_path_str:
                return
            configuration = ProjectConfiguration()
            wait(configuration.write(Path(configuration_file_path_str)))
            project_window = ProjectWindow(self._app)
            project_window.show()
            self.close()

    def _demo(self) -> None:
        with ExceptionCatcher(self):
            serve_window = ServeDemoWindow(self._app, parent=self)
            serve_window.show()

    @sync
    async def clear_caches(self) -> None:
        async with ExceptionCatcher(self):
            await self._app.cache.clear()

    def open_application_configuration(self) -> None:
        with ExceptionCatcher(self):
            window = ApplicationConfiguration(self._app, parent=self)
            window.show()


class _WelcomeText(Text):
    pass


class _WelcomeTitle(_WelcomeText):
    pass


class _WelcomeHeading(_WelcomeText):
    pass


class _WelcomeAction(QPushButton):
    pass


class WelcomeWindow(BettyPrimaryWindow):
    # Allow the window to be as narrow as it can be.
    window_width = 1
    # This is a best guess at the minimum required height, because if we set this to 1, like the width, some of the
    # text will be clipped.
    window_height = 600

    def __init__(
        self,
        app: App,
    ):
        super().__init__(app)

        central_layout = QVBoxLayout()
        central_layout.addStretch()
        central_widget = QWidget()
        central_widget.setLayout(central_layout)
        self.setCentralWidget(central_widget)

        self._welcome = _WelcomeTitle()
        self._welcome.setAlignment(Qt.AlignmentFlag.AlignCenter)
        central_layout.addWidget(self._welcome)

        self._welcome_caption = _WelcomeText()
        central_layout.addWidget(self._welcome_caption)

        self._project_instruction = _WelcomeHeading()
        self._project_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        central_layout.addWidget(self._project_instruction)

        project_layout = QHBoxLayout()
        central_layout.addLayout(project_layout)

        self.open_project_button = _WelcomeAction(self)
        self.open_project_button.released.connect(self.open_project)
        project_layout.addWidget(self.open_project_button)

        self.new_project_button = _WelcomeAction(self)
        self.new_project_button.released.connect(self.new_project)
        project_layout.addWidget(self.new_project_button)

        self._demo_instruction = _WelcomeHeading()
        self._demo_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        central_layout.addWidget(self._demo_instruction)

        self.demo_button = _WelcomeAction(self)
        self.demo_button.released.connect(self._demo)
        central_layout.addWidget(self.demo_button)

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._welcome.setText(self._app.localizer._('Welcome to Betty'))
        self._welcome_caption.setText(self._app.localizer._('Betty helps you visualize and publish your family history by building interactive genealogy websites out of your <a href="{gramps_url}">Gramps</a> and <a href="{gedcom_url}">GEDCOM</a> family trees.').format(
            gramps_url='https://gramps-project.org/',
            gedcom_url='https://en.wikipedia.org/wiki/GEDCOM',
        ))
        self._project_instruction.setText(self._app.localizer._('Work on a new or existing site of your own'))
        self.open_project_button.setText(self._app.localizer._('Open an existing project'))
        self.new_project_button.setText(self._app.localizer._('Create a new project'))
        self._demo_instruction.setText(self._app.localizer._('View a demonstration of what a Betty site looks like'))
        self.demo_button.setText(self._app.localizer._('View a demo site'))


class _AboutBettyWindow(BettyMainWindow):
    window_width = 500
    window_height = 100

    def __init__(
        self,
        app: App,
        *,
        parent: QObject | None = None,
    ):
        super().__init__(app, parent=parent)
        self._label = Text()
        self._label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setCentralWidget(self._label)

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._label.setText(''.join(map(lambda x: '<p>%s</p>' % x, [
            self._app.localizer._('Version: {version}').format(
                version=wait(about.version_label()),
            ),
            self._app.localizer._('Copyright 2019-{year} Bart Feenstra & contributors. Betty is made available to you under the <a href="https://www.gnu.org/licenses/gpl-3.0.en.html">GNU General Public License, Version 3</a> (GPLv3).').format(
                year=datetime.now().year,
            ),
            self._app.localizer._('Follow Betty on <a href="https://twitter.com/Betty_Project">Twitter</a> and <a href="https://github.com/bartfeenstra/betty">Github</a>.'),
        ])))

    @property
    def window_title(self) -> Localizable:
        return Str._('About Betty')


class ApplicationConfiguration(BettyMainWindow):
    window_width = 400
    window_height = 150

    def __init__(
        self,
        app: App,
        *,
        parent: QObject | None = None,
    ):
        super().__init__(app, parent=parent)

        self._form = QFormLayout()
        form_widget = QWidget()
        form_widget.setLayout(self._form)
        self.setCentralWidget(form_widget)
        self._locale_collector = TranslationsLocaleCollector(self._app, set(self._app.localizers.locales))
        for row in self._locale_collector.rows:
            self._form.addRow(*row)

    @property
    def window_title(self) -> Localizable:
        return Str._('Configuration')



"""
Provide a serialization API.
"""
from __future__ import annotations

from typing import TypeVar, Sequence, Mapping, overload, Literal, TypeAlias, Any

T = TypeVar('T')
U = TypeVar('U')


class Void:
    pass


DumpType: TypeAlias = bool | int | float | str | None | list['Dump'] | dict[str, 'Dump']
DumpTypeT = TypeVar('DumpTypeT', bound=DumpType)

Dump: TypeAlias = bool | int | float | str | None | Sequence['Dump'] | Mapping[str, 'Dump']
DumpT = TypeVar('DumpT', bound=Dump)
DumpU = TypeVar('DumpU', bound=Dump)

VoidableDump: TypeAlias = Dump | type[Void]
VoidableDumpT = TypeVar('VoidableDumpT', bound=VoidableDump)
VoidableDumpU = TypeVar('VoidableDumpU', bound=VoidableDump)

ListDump: TypeAlias = list[DumpT]

DictDump: TypeAlias = dict[str, DumpT]

VoidableListDump: TypeAlias = list[VoidableDumpT]

VoidableDictDump: TypeAlias = dict[str, VoidableDumpT]

_MinimizableDump: TypeAlias = VoidableDump | VoidableListDump[VoidableDumpT] | VoidableDictDump[VoidableDumpT]


@overload
def minimize(dump: _MinimizableDump[VoidableDump], voidable: Literal[True] = True) -> VoidableDump:
    pass


@overload
def minimize(dump: _MinimizableDump[VoidableDump], voidable: Literal[False]) -> Dump:
    pass


def minimize(dump: _MinimizableDump[VoidableDump], voidable: bool = True) -> VoidableDump:
    """
    Minimize a configuration dump by removing any Void configurationfrom sequences and mappings.
    """
    if isinstance(dump, (Sequence, Mapping)) and not isinstance(dump, str):
        if isinstance(dump, Sequence):
            dump = [
                value
                for value
                in dump
                if value is not Void
            ]
            for key in reversed(range(len(dump))):
                if dump[key] is Void:
                    del dump[key]
        if isinstance(dump, Mapping):
            dump = {
                key: value
                for key, value
                in dump.items()
                if value is not Void
            }
        if len(dump) or not voidable:
            return dump  # type: ignore[return-value]
        return Void
    return dump


def void_none(value: VoidableDump) -> VoidableDump:
    """
    Passthrough a value, but convert Void to None.
    """
    return Void if value is None else value


def none_void(value: VoidableDump) -> VoidableDump:
    """
    Passthrough a value, but convert None to Void.
    """
    return None if value is Void else value


@overload
def dump_default(dump: DictDump[Dump], key: str, default_type: type[dict[Any, Any]]) -> DictDump[Dump]:
    pass


@overload
def dump_default(dump: DictDump[Dump], key: str, default_type: type[list[Any]]) -> ListDump[Dump]:
    pass


def dump_default(dump, key, default_type):
    """
    Add a key and value to a dump, if the key does not exist yet.
    """
    try:
        assert isinstance(dump[key], default_type)
    except KeyError:
        dump[key] = default_type()
    return dump[key]  # type: ignore[return-value]


class Dumpable:
    def dump(self) -> VoidableDump:
        """
        Dump this instance to a portable format.
        """
        raise NotImplementedError(repr(self))



"""
Provide JSON utilities.
"""
from __future__ import annotations

from typing import Any, TypeVar, TYPE_CHECKING, cast

from jsonschema.validators import Draft202012Validator
from referencing import Resource, Registry

from betty.serde.dump import DictDump, Dump, dump_default
from betty.string import upper_camel_case_to_lower_camel_case

if TYPE_CHECKING:
    from betty.app import App


T = TypeVar('T')


class Schema:
    def __init__(self, app: App):
        self._app = app

    async def build(self) -> DictDump[Dump]:
        from betty.model import get_entity_type_name

        schema: DictDump[Dump] = {
            '$schema': 'https://json-schema.org/draft/2020-12/schema',
            '$id': self._app.static_url_generator.generate('schema.json', absolute=True),
        }

        definitions = dump_default(schema, 'definitions', dict)
        entity_definitions = dump_default(definitions, 'entity', dict)
        response_definitions = dump_default(definitions, 'response', dict)

        # Add entity schemas.
        for entity_type in self._app.entity_types:
            entity_type_schema_name = upper_camel_case_to_lower_camel_case(get_entity_type_name(entity_type))
            entity_type_schema = await entity_type.linked_data_schema(self._app)
            entity_type_schema_definitions = cast(DictDump[Dump], entity_type_schema.pop('definitions', {}))
            for definition_name, definition_schema in entity_type_schema_definitions.items():
                if definition_name not in definitions:
                    definitions[definition_name] = definition_schema
            entity_definitions[entity_type_schema_name] = entity_type_schema
            entity_definitions[f'{entity_type_schema_name}Collection'] = {
                'type': 'array',
                'items': {
                    'type': 'string',
                    'format': 'uri',
                },
            }
            response_definitions[f'{entity_type_schema_name}Collection'] = {
                'type': 'object',
                'properties': {
                    'collection': {
                        '$ref': f'#/definitions/entity/{entity_type_schema_name}Collection',
                    },
                },
            }

        # Add the HTTP error response.
        response_definitions['error'] = {
            'type': 'object',
            'properties': {
                '$schema': ref_json_schema(schema),
                'message': {
                    'type': 'string',
                },
            },
            'required': [
                '$schema',
                'message',
            ],
            'additionalProperties': False,
        }

        return schema

    async def validate(self, data: Any) -> None:
        """
        Validate JSON against the Betty JSON schema.
        """
        schema = Resource.from_contents(await self.build())
        schema_registry = schema @ Registry()  # type: ignore[operator, var-annotated]
        validator = Draft202012Validator(
            {
                '$ref': data['$schema'],
            },
            registry=schema_registry,
        )
        validator.validate(data)


def add_property(schema: DictDump[Dump], property_name: str, property_schema: DictDump[Dump], property_required: bool = True) -> None:
    """
    Add a property to an object schema.
    """
    schema_properties = dump_default(schema, 'properties', dict)
    schema_properties[property_name] = property_schema
    if property_required:
        schema_required = dump_default(schema, 'required', list)
        schema_required.append(property_name)


def ref_locale(root_schema: DictDump[Dump]) -> DictDump[Dump]:
    """
    Reference the locale schema.
    """
    definitions = dump_default(root_schema, 'definitions', dict)
    if 'locale' not in definitions:
        definitions['locale'] = {
            'type': 'string',
            'description': 'A BCP 47 locale identifier (https://www.ietf.org/rfc/bcp/bcp47.txt).',
        }
    return {
        '$ref': '#/definitions/locale',
    }


def ref_json_schema(root_schema: DictDump[Dump]) -> DictDump[Dump]:
    """
    Reference the JSON Schema schema.
    """
    definitions = dump_default(root_schema, 'definitions', dict)
    if 'schema' not in definitions:
        definitions['schema'] = {
            'type': 'string',
            'format': 'uri',
            'description': 'A JSON Schema URI.',
        }
    return {
        '$ref': '#/definitions/schema',
    }



"""Provide the Graphical User Interface (GUI) for Betty Desktop."""
from __future__ import annotations

import pickle
from typing import Any, TypeVar

from PyQt6.QtCore import pyqtSlot, QObject
from PyQt6.QtGui import QPalette
from PyQt6.QtWidgets import QApplication, QWidget

from betty.app import App
from betty.gui.error import ExceptionError, _UnexpectedExceptionError
from betty.locale import Str
from betty.serde.format import FormatRepository

QWidgetT = TypeVar('QWidgetT', bound=QWidget)


def get_configuration_file_filter() -> Str:
    """
    Get the Qt file filter for project configuration files.
    """
    formats = FormatRepository()
    return Str._(
        'Betty project configuration ({supported_formats})',
        supported_formats=' '.join(
            f'*.{extension}'
            for format
            in formats.formats
            for extension in format.extensions
        ),
    )


class GuiBuilder:
    def gui_build(self) -> QWidget:
        raise NotImplementedError(repr(self))


def mark_valid(widget: QWidget) -> None:
    """
    Mark a widget as currently containing valid input.
    """
    widget.setProperty('invalid', 'false')
    widget.setStyle(widget.style())
    widget.setToolTip('')


def mark_invalid(widget: QWidget, reason: str) -> None:
    """
    Mark a widget as currently containing invalid input.
    """
    widget.setProperty('invalid', 'true')
    widget.setStyle(widget.style())
    widget.setToolTip(reason)


class BettyApplication(QApplication):
    def _is_dark_mode(self) -> bool:
        palette = self.palette()
        window_lightness = palette.color(QPalette.ColorRole.Window).lightness()
        window_text_lightness = palette.color(QPalette.ColorRole.WindowText).lightness()
        return window_lightness < window_text_lightness

    def _stylesheet(self) -> str:
        if self._is_dark_mode():
            caption_color = '#eeeeee'
        else:
            caption_color = '#333333'
        return f"""
            Caption {{
                color: {caption_color};
                font-size: 14px;
                margin-bottom: 0.3em;
            }}

            Code {{
                font-family: monospace;
            }}

            QLineEdit[invalid="true"] {{
                border: 1px solid red;
                color: red;
            }}

            QPushButton[pane-selector="true"] {{
                padding: 10px;
            }}

            LogRecord[level="50"],
            LogRecord[level="40"] {{
                color: red;
            }}

            LogRecord[level="30"] {{
                color: yellow;
            }}

            LogRecord[level="20"] {{
                color: green;
            }}

            LogRecord[level="10"],
            LogRecord[level="0"] {{
                color: white;
            }}

            _WelcomeText {{
                padding: 10px;
            }}

            _WelcomeTitle {{
                font-size: 20px;
                padding: 10px;
            }}

            _WelcomeHeading {{
                font-size: 16px;
                margin-top: 50px;
            }}

            _WelcomeAction {{
                padding: 10px;
            }}
            """

    def __init__(self, *args: Any, app: App, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.setApplicationName('Betty')
        self.setStyleSheet(self._stylesheet())
        self._app = app

    @pyqtSlot(
        type,
        bytes,
        QObject,
        bool,
    )
    def _show_user_facing_error(
        self,
        error_type: type[Exception],
        pickled_error_message: bytes,
        parent: QObject,
        close_parent: bool,
    ) -> None:
        error_message = pickle.loads(pickled_error_message)
        window = ExceptionError(self._app, error_message, error_type, parent=parent, close_parent=close_parent)
        window.show()

    @pyqtSlot(
        type,
        str,
        str,
        QObject,
        bool,
    )
    def _show_unexpected_exception(
        self,
        error_type: type[Exception],
        error_message: str,
        error_traceback: str,
        parent: QObject,
        close_parent: bool,
    ) -> None:
        window = _UnexpectedExceptionError(self._app, error_type, error_message, error_traceback, parent=parent, close_parent=close_parent)
        window.show()
